import os

import cospar as cs
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import skbio
from ete3 import Tree
from matplotlib import pyplot as plt
from matplotlib import rcParams
from skbio import DistanceMatrix

from . import metric, plotting, similarity
from .settings import *


def methyltree_core(
    adata,
    out_dir,
    save_data_des,
    clone_key,
    compute_similarity=False,
    similarity_method="correlation",
    similarity_correction=True,
    similarity_correction_method="fast",
    similarity_correction_step_size=0.05,
    similarity_normalize=True,
    remove_celltype_signal=False,
    remove_celltype_signal_method="fast",
    remove_celltype_signal_outlier_thresh=1,
    cell_type_key=None,
    neighbor_info=None,
    heatmap_ordering_method="UPGMA",
    optimize_tree=False,
    optimize_background_cutoff=0.8,
    optimize_show_legend=True,
    permute_cell_order=False,
    selected_idx=None,
    fig_out=True,
):
    """
    THis core algorithm accept an adata, and performs the following:
    1, compute the similarity matrix
    2, similarity correction
    3, similarity zscore
    4, Build the lineage with NJ method
    """

    os.makedirs(out_dir, exist_ok=True)
    if (
        (cell_type_key is None)
        and (neighbor_info is None)
        and (remove_celltype_signal == True)
    ):
        raise ValueError(
            "Missing cell_type_key or neighbor_info for removing celltype signal"
        )

    clone_id_list = np.array(adata.obs[clone_key]).astype(str)
    clone_id_list = clone_id_list[(clone_id_list != "nan") & (clone_id_list != "NA")]
    if len(set(clone_id_list)) == len(clone_id_list):
        raise ValueError(
            f"clone_key {clone_key} does not have clones with >1 cells. Change clone_key please!"
        )

    ##########################################
    ## check or update similarity matrix
    ##########################################

    raw_similarity_key = f"X_similarity_{similarity_method}_raw"

    if raw_similarity_key not in adata.obsm.keys():
        compute_similarity = True
        print(f"{raw_similarity_key} not found in adata.obsm")
    else:
        X_shape = adata.obsm[raw_similarity_key].shape
        if X_shape[0] != X_shape[1]:
            print(f"Similarity shape mismatch ({X_shape})")
            compute_similarity = True

    if compute_similarity:
        print("re-compute similarity matrix")
        X_similarity, shared_site_matrix = similarity.compute_similarity_matrix(
            adata.X,
            method=similarity_method,
        )
        adata.obsm[raw_similarity_key] = X_similarity
        adata.obsm["X_shared_sites"] = shared_site_matrix
    adata.obsm["X_similarity"] = adata.obsm[raw_similarity_key].copy()

    # permute the adata so that our predicted ordering will not be influenced by the original ordering of the cell
    if permute_cell_order and (neighbor_info is None):
        print("permute ordering")
        permut_idx = np.random.default_rng(seed=42).permutation(adata.shape[0])
        adata = adata[permut_idx]
        adata.obsm["X_similarity"] = adata.obsm["X_similarity"][:, permut_idx]
        if "X_join_N" in adata.obsm.keys():
            adata.obsm["X_join_N"] = adata.obsm["X_join_N"][:, permut_idx]

    ###################################################
    ##   X_similarity regression to remove noise
    ###################################################

    if similarity_correction:
        # the initial rescaling helps to avoid negative correlations
        if similarity_normalize:
            X_similarity_old = similarity.rescale_similarity_matrix(
                adata.obsm["X_similarity"].copy()
            )
        else:
            X_similarity_old = adata.obsm["X_similarity"].copy()
        for _ in range(1):
            currect_epsilon = np.max([similarity_correction_step_size - 0.01 * _, 0.01])
            print(
                f"correct similarity: outer loop {_};  current epsilon {currect_epsilon}"
            )
            X_similarity_new = similarity.correct_similarity_matrix(
                X_similarity_old,
                fig_out=fig_out,
                epsilon=currect_epsilon,
                method=similarity_correction_method,
            )
            tmp = X_similarity_new - X_similarity_old
            tmp_std = np.std((tmp).flatten())
            print(f"std: {tmp_std:.3f}")
            X_similarity_old = X_similarity_new
            if tmp_std < 0.005:
                break

        adata.obsm["X_similarity"] = X_similarity_new

    ###################################################
    ##   X_similarity z-score to enhance lineage signal
    ###################################################
    # generate cell_type_key
    if cell_type_key is not None:
        # get rid of na or NAN, which has no cell type annotation
        adata.obs[cell_type_key] = adata.obs[cell_type_key].astype(str)
        sel_idx = ~adata.obs[cell_type_key].isin(["nan", "NA"])
        if (
            (np.sum(~sel_idx) > 0)
            and (neighbor_info is None)
            and remove_celltype_signal
        ):
            print("Warning: exclude cells with no type annotation")
            adata = adata[sel_idx]
            sel_idx = np.array(sel_idx)
            adata.obsm["X_similarity"] = adata.obsm["X_similarity"][:, sel_idx]
            print(f"New adata shape: {adata.shape}")
        cell_type_array = np.array(adata.obs[cell_type_key]).astype(str)
    else:
        cell_type_array = None

    if remove_celltype_signal:
        if (remove_celltype_signal_method == "fast") and (neighbor_info is None):
            # This method is much faster
            print("Remove cell-type signal with fast method")
            zscore_X_similarity = similarity.remove_celltype_signal_discrete_state(
                adata.obsm["X_similarity"],
                cell_type_array=cell_type_array,
                outlier_threshold=remove_celltype_signal_outlier_thresh,
            )
        else:
            if neighbor_info is None:
                neighbor_info = []
                for j, x in enumerate(cell_type_array):
                    neighbor_IDs = np.nonzero(cell_type_array == x)[0]
                    if len(neighbor_IDs) == 0:
                        neighbor_IDs = np.arange(
                            len(cell_type_array)
                        )  # if NaN, use the entire array as the background
                    neighbor_info.append(neighbor_IDs)

            # This method is very slow
            print("Remove cell-type signal with slow/original method")
            zscore_X_similarity = similarity.remove_celltype_signal_continuous_state(
                adata.obsm["X_similarity"],
                neighbor_info=neighbor_info,
                outlier_threshold=remove_celltype_signal_outlier_thresh,
            )

        adata.obsm["X_similarity"] = zscore_X_similarity

    #############################################################
    ##   show similarity histogram going for downstream analysis
    #############################################################

    if similarity_normalize:
        print("similarity normalize----")
        X_similarity = similarity.rescale_similarity_matrix(
            adata.obsm["X_similarity"].copy()
        )
    else:
        X_similarity = adata.obsm["X_similarity"].copy()

    if selected_idx is not None:
        X_similarity = X_similarity[selected_idx][:, selected_idx]
        adata = adata[selected_idx]

    adata.obsm["X_similarity"] = X_similarity

    if fig_out:
        fig, ax = plt.subplots(figsize=(4, 3))
        tmp_data = (
            np.array(X_similarity[np.triu_indices_from(X_similarity, k=1)]) + 0.01
        )  # 0.01 is pseudocount to avoid zero when taking log in the following histplot
        sns.histplot(tmp_data)
        plt.yscale("log")
        plt.xlabel("Similarity")
        plt.show()

    ###################################################
    ## Lineage tree inference or cell state ordering
    ###################################################
    Allowed_methods = ["NJ", "HierarchicalClustering", "UPGMA", "FASTME"]

    ### Tree reconstruction
    print(f"Reconstruction method: {heatmap_ordering_method}")
    tree_path = f"{out_dir}/lineage_tree_{clone_key}_{save_data_des}.txt"
    if heatmap_ordering_method in ["NJ", "UPGMA", "FASTME"]:
        dissimilarity = np.max(X_similarity) - X_similarity
        dissimilarity = dissimilarity - np.min(dissimilarity)
        for j in range(X_similarity.shape[0]):
            dissimilarity[j, j] = 0
        dissimilarity = dissimilarity / np.max(dissimilarity)
        dm = DistanceMatrix(
            dissimilarity,
            [f"{x}" for x in list(np.arange(dissimilarity.shape[0]).astype(str))],
        )

        if heatmap_ordering_method == "NJ":
            my_tree = skbio.tree.nj(dm)
            # my_tree = lineage_tree.nj(dm, method=NJ_parameter)
            my_tree.write(tree_path)
        elif heatmap_ordering_method == "FASTME":
            f_out = run_fastme(dissimilarity, out_dir)
            my_tree = skbio.TreeNode.read(f_out)
            my_tree.write(tree_path)
        elif heatmap_ordering_method == "UPGMA":
            from biotite.sequence.phylo import upgma

            my_tree = upgma(dissimilarity).to_newick()
            with open(tree_path, "w") as f:
                f.write(my_tree)

        with open(tree_path, "r") as f:
            my_tree = Tree(f.read())

        ######################################################
        ## Improve lineage tree with stochastic optimization
        ######################################################
        if optimize_tree:
            print("Optimize lineage tree")
            my_tree = optimize_lineage_tree(
                adata,
                my_tree,
                background_cutoff=optimize_background_cutoff,
                clone_key=clone_key,
                plot_heatmap=False,
                show_legend=optimize_show_legend,
            )

        order_x = np.array(my_tree.get_leaf_names()).astype(int)

    elif heatmap_ordering_method == "HierarchicalClustering":
        from scipy.cluster.hierarchy import leaves_list, linkage, optimal_leaf_ordering
        from scipy.spatial.distance import pdist, squareform

        # Assume cov_matrix is your covariance matrix
        # Convert covariance to dissimilarity
        dissimilarity = np.max(X_similarity) - X_similarity
        # Convert to condensed distance matrix
        dist_mat = squareform(dissimilarity)
        # Compute linkage matrix
        Z = linkage(dist_mat, method="average")
        # Get the optimal leaf ordering
        order_x = leaves_list(optimal_leaf_ordering(Z, dist_mat))
    else:
        raise ValueError(f"method must be among: {Allowed_methods}")
    return adata, order_x


def bootstrap_lineage_tree(
    adata_orig,
    out_dir,
    save_data_des,
    clone_key,
    cell_type_key=None,
    selected_idx=None,
    num_iterations=100,
    sampled_replacement=False,
    sample_fraction=0.8,
    similarity_correction=False,
    similarity_normalize=True,
    remove_celltype_signal=False,
    fig_dir=None,
    exact_match=False,
    recompute=False,
    heatmap_ordering_method="UPGMA",
    cores=64,
):
    """
    Generate tree support
    """

    if not os.path.exists(out_dir):
        os.makedirs(out_dir, exist_ok=True)

    if fig_dir is None:
        fig_dir = f"{out_dir}/figure"
    sc.set_figure_params()
    sc.settings.figdir = fig_dir

    num_cols_feature = adata_orig.shape[1]
    tree_dict = {}  ## a dictionary for appearance counts of all observed sub-trees
    node_support = {}  ## a dictionary for support values of all observed sub-trees

    parameter_str = f"{heatmap_ordering_method}_frac{sample_fraction}_normalize{similarity_normalize}_correction{similarity_correction}_celltypesignal{remove_celltype_signal}_celltypekey{cell_type_key}"

    ###################################################
    ############## tree from bootstrap
    ###################################################
    def compute_sub_sampled_adata(round_num):
        output_tree_dir = f"{out_dir}/lineage_tree_{clone_key}_{save_data_des}_bootstrap_{round_num}_{parameter_str}.txt"
        if (not os.path.exists(output_tree_dir)) or recompute:
            print(f"num_iterations : {round_num}")
            sample_size = int(sample_fraction * num_cols_feature)

            if sampled_replacement:
                sampled_cols = np.random.choice(
                    num_cols_feature, size=sample_size, replace=True
                )
            else:
                sampled_cols = np.random.choice(
                    num_cols_feature, size=sample_size, replace=False
                )

            adata_bootstrap = adata_orig[:, sorted(sampled_cols)].copy()

            print("adata shape: ", adata_bootstrap.shape)

            # the output of the next step is stored here
            adata_bootstrap, _ = methyltree_core(
                adata_bootstrap,
                out_dir,
                f"{save_data_des}_bootstrap_{round_num}_{parameter_str}",
                clone_key,
                selected_idx=selected_idx,
                fig_out=False,  # to avoid figure printing
                compute_similarity=True,  # force computation of the similarity
                cell_type_key=cell_type_key,
                similarity_correction=similarity_correction,
                similarity_normalize=similarity_normalize,
                remove_celltype_signal=remove_celltype_signal,
            )

    if cores > num_iterations:
        cores = num_iterations

    from joblib import Parallel, delayed

    results = Parallel(n_jobs=cores)(
        delayed(compute_sub_sampled_adata)(i) for i in range(num_iterations)
    )

    for round_num in range(num_iterations):
        output_tree_dir = f"{out_dir}/lineage_tree_{clone_key}_{save_data_des}_bootstrap_{round_num}_{parameter_str}.txt"

        ###################################################
        ### update sub-tree frequency based on current tree
        ###################################################

        ## read bootstrapped tree and compute the frequency of each sub-tree
        with open(output_tree_dir, "r") as f:
            tree_str = f.read().strip()
        tree = Tree(newick=tree_str, format=1)
        # tree.sort_descendants()  ## the original label 0,1,2,..., is abitrary. We should not sort the leaves by this order.
        ## Also, using our fuzzy match method, the result is no longer sensitive to the order of leaf are not longer

        ##  set the tree branch length to be 1 to avoid problems in exact match
        for node in tree.traverse():
            node.dist = 1

        ## update sub-tree freqeuncy
        for node in tree.traverse("postorder"):
            if not node.is_leaf():
                if exact_match:
                    ## when the same set of cells appear in this sub-tree, and the order is also the same
                    subtree_str = node.write()
                else:
                    ## when the same set of cells appear in this sub-tree, but the order can be different
                    subtree_str = ",".join(
                        np.array(
                            sorted(np.array(node.get_leaf_names()).astype(int))
                        ).astype(str)
                    )

                if subtree_str not in tree_dict:
                    tree_dict[subtree_str] = 1
                else:
                    tree_dict[subtree_str] += 1

    ## compute the support value for each sub-tree
    for node in tree_dict:
        node_support[node] = tree_dict[node] / num_iterations

    ###################################################
    ############## original-tree generation
    ###################################################

    orig_tree_dir = (
        f"{out_dir}/lineage_tree_{clone_key}_{save_data_des}_{parameter_str}.txt"
    )
    if (not os.path.exists(orig_tree_dir)) or recompute:
        adata_orig, _ = methyltree_core(
            adata_orig,
            out_dir,
            f"{save_data_des}_{parameter_str}",
            clone_key,
            selected_idx=selected_idx,
            fig_out=False,  # to avoid figure printing
            cell_type_key=cell_type_key,
            similarity_correction=similarity_correction,
            remove_celltype_signal=remove_celltype_signal,
            heatmap_ordering_method=heatmap_ordering_method,
            similarity_normalize=similarity_normalize,
        )

    ############################################
    ### get dictory for updating leaf names
    ############################################

    if selected_idx is not None:
        adata_new = adata_orig[selected_idx].copy()
    else:
        adata_new = adata_orig

    df_state_clone = pd.DataFrame(
        {
            "state_info": np.arange(adata_new.shape[0]).astype(str),
            "clone_id": adata_new.obs[clone_key].to_list(),
        }
    ).fillna("nan")

    df_state_clone.index = df_state_clone["state_info"].to_list()
    clone_name_map = dict(
        zip(df_state_clone["state_info"].astype(str), df_state_clone["clone_id"])
    )
    if cell_type_key is not None:
        df_type = pd.DataFrame(
            {cell_type_key: adata_new.obs[cell_type_key].to_list()}
        ).reset_index()
        cell_type_map = dict(zip(df_type.index.astype(str), df_type[cell_type_key]))

    ##########################################
    ### update support values and save results
    ##########################################

    with open(orig_tree_dir, "r") as f:
        tree_str_tmp = f.read().strip()
    tmp_tree_support = Tree(newick=tree_str_tmp, format=1)
    # tmp_tree_support.sort_descendants()
    ## set the branch length to be uniform, for better visualization
    for node in tmp_tree_support.traverse():
        node.dist = 1

    for node in tmp_tree_support.traverse("postorder"):
        if not node.is_leaf():
            if exact_match:
                subtree_str = node.write()
            else:
                subtree_str = ",".join(
                    np.array(
                        sorted(np.array(node.get_leaf_names()).astype(int))
                    ).astype(str)
                )

            # set the name to be the support value
            if subtree_str in node_support:
                node.name = f"{node_support[subtree_str]:.2f}"
            else:
                node.name = 0

    for j, leaf in enumerate(tmp_tree_support.iter_leaves()):
        if (cell_type_key is None) or (cell_type_key == clone_key):
            new_name = f"{leaf.name}" + "-" + str(clone_name_map[leaf.name])
        else:
            new_name = (
                f"{leaf.name}"
                + "-"
                + str(cell_type_map[leaf.name])
                + "-"
                + clone_name_map[leaf.name]
            )
        leaf.name = new_name

    tmp_tree_support.write(
        format=1,
        outfile=f"{out_dir}/lineage_tree_{clone_key}_{save_data_des}_{heatmap_ordering_method}_support.txt",
    )

    return tmp_tree_support


def run_fastme(dissimilarity, out_dir):

    # Write the distance matrix to a temporary file
    input_file = os.path.join(out_dir, "fastme_tmp.dist")
    precision = 100000
    dissimilarity = (dissimilarity * precision).astype(int) / precision

    name_list = [f"{x}" for x in list(np.arange(dissimilarity.shape[0]).astype(str))]

    with open(input_file, "w") as file:
        file.write(f"{len(name_list)}\n")
        from tqdm import tqdm

        for j, x in tqdm(enumerate(list(dissimilarity))):
            data = "\t".join(x.astype(str))
            file.write(str(name_list[j]) + "\t" + "".join(data) + "\n")

    # Define the output file path
    f_out = os.path.join(out_dir, "fastme_tmp.nwk")

    # Construct the command
    command = f"fastme -i {input_file} -o {f_out} -n -s"

    # Run the command
    exit_code = os.system(command)

    if exit_code != 0:
        raise RuntimeError(f"FastME did not run successfully. Exit code: {exit_code}")

    return f_out


#############################################
## lineage tree operation and optimization
#############################################


def optimize_lineage_tree(
    adata,
    my_tree,
    background_cutoff,
    clone_key,
    iterations=10000,
    plot_heatmap=True,
    plot_history=True,
    show_legend=True,
):
    X_similarity = similarity.rescale_similarity_matrix(
        adata.obsm["X_similarity"].copy()
    )
    order_x = np.array(my_tree.get_leaf_names()).astype(int)

    ## determine clone size scale
    for i in range(X_similarity.shape[0]):
        X_similarity[i, i] = 0
    high_idx = np.nonzero(X_similarity[order_x][:, order_x] > background_cutoff)

    temp = abs(high_idx[0] - high_idx[1])
    clone_size_scale = np.median(temp)
    print(f"Clone size scale: {clone_size_scale}")

    ## define the weight matrix
    for i in range(X_similarity.shape[0]):
        X_similarity[i, i] = 1

    factor = 2
    size = X_similarity.shape[0]
    weight_matrix = np.zeros((size, size))
    for i in range(weight_matrix.shape[0]):
        for j in range(weight_matrix.shape[0]):
            if abs(i - j) > factor * clone_size_scale:
                weight_matrix[i, j] = (
                    np.exp(-factor) * (factor * clone_size_scale) / abs(i - j)
                )  # np.exp(-factor)*(weight_matrix.shape[0]-abs(i-j)+factor*clone_size_scale)/
            else:
                weight_matrix[i, j] = np.exp(-abs(i - j) / clone_size_scale)
            # weight_matrix[i,j]=np.exp(-abs(i-j)/clone_size_scale)

    ## perform optimization
    current_score = np.sum(X_similarity[order_x][:, order_x] * weight_matrix)
    result = []
    from tqdm import tqdm

    for j in tqdm(range(iterations)):
        # my_tree_new=omic_ana.random_subtree_pruning_and_regrafting(my_tree)

        my_tree_new = random_subtree_swapping(my_tree)

        order_x_new = np.array(my_tree_new.get_leaf_names()).astype(int)
        next_score = np.sum(X_similarity[order_x_new][:, order_x_new] * weight_matrix)
        # print(next_score-current_score)

        if next_score > current_score:
            my_tree = my_tree_new
            current_score = next_score

            ordered_clone_array = adata.obs[clone_key].astype(str)[order_x_new]
            unique_clone_set = list(set(ordered_clone_array))
            for x0 in unique_clone_set:
                if np.sum(ordered_clone_array == x0) == 1:
                    ordered_clone_array[ordered_clone_array == x0] = "nan"

            df_accuracy = metric.lineage_accuracy_from_leaf_order(ordered_clone_array)
            mean_accuracy = df_accuracy[df_accuracy["clone_size"] > 1].mean()
            mean_accuracy = mean_accuracy[
                ["accuracy", "entropy", "continuity", "wassertein"]
            ]
            print(f"iter-{j}; score {current_score}; continuity {mean_accuracy[2]}")

            result.append(
                [
                    current_score,
                    mean_accuracy[0],
                    mean_accuracy[1],
                    mean_accuracy[2],
                    mean_accuracy[3],
                ]
            )

    ## plot the results
    if plot_history:
        plt.clf()
        cs.settings.set_figure_params()
        df_tmp = pd.DataFrame(result)
        fig, ax = plt.subplots()
        ax.scatter(df_tmp[0], df_tmp[1], label="accuracy")
        ax.scatter(df_tmp[0], df_tmp[2], label="entropy")
        ax.scatter(df_tmp[0], df_tmp[3], label="continuity")
        ax.scatter(df_tmp[0], df_tmp[4], label="wassertein")
        plt.xlabel("Optimization score")
        plt.ylabel("Accuracy")
        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))

    if plot_heatmap:
        vmin = np.percentile(X_similarity.flatten(), 50)
        vmax = np.max(np.triu(X_similarity, k=1))

        plotting.plot_similarity_heatmap_with_clone_label(
            X_similarity[order_x_new][:, order_x_new],
            label=adata.obs[clone_key].astype(str)[order_x_new],
            vmin=vmin,
            vmax=vmax,
            color_bar=False,
            save_name=f"heatmap.pdf",
            # title=title,  # f"{stage}_{annotation}",
            # clone_color_dict=clone_color_dict,
            # show_label=show_label,
            cell_index=None,
            show_legend=show_legend,
        )

    return my_tree


def is_subtree(subtree, tree):
    return subtree.write(format=5)[:-1] in tree.write(format=5)[:-1]


def swap_branches(tree, branch_1, branch_2):
    X = branch_1.write(format=5)[:-1]
    Y = branch_2.write(format=5)[:-1]
    Orig = tree.write(format=5)
    seq_list = Orig.split(X)
    if Y in seq_list[0]:
        final = Y.join([X.join(seq_list[0].split(Y)), seq_list[1]])
    else:
        final = Y.join([seq_list[0], X.join(seq_list[1].split(Y))])
    return Tree(final)


def swap_descendants(tree, descendant1, descendant2):
    # does not add the sub branch_1 to the exactly the same location of branch_2
    up_1 = descendant1.up
    up_2 = descendant2.up
    down_1 = descendant1.detach()
    down_2 = descendant2.detach()
    up_1.add_child(down_2)
    up_2.add_child(down_1)
    return tree


# Define a function to perform random subtree pruning and regrafting
def random_subtree_pruning_and_regrafting(tree_orig):
    tree = tree_orig.copy()
    import random

    # Select a random node for subtree pruning
    branch_1 = random.choice(tree.get_descendants())

    # Select a random target node for regrafting
    # Ensure the target node is not a descendant of the pruned subtree
    while True:
        branch_2 = random.choice(tree.get_descendants())
        if not (is_subtree(branch_2, branch_1) or is_subtree(branch_1, branch_2)):
            break

    # print(branch_1)
    # print(branch_2)

    # return swap_branches(tree, branch_1, branch_2)
    return swap_descendants(tree, branch_1, branch_2)


def random_subtree_swapping(tree_orig):
    tree = tree_orig.copy()
    import random

    # Select a random node for subtree pruning
    branch_1 = random.choice(tree.get_descendants())
    branch_1.swap_children()
    return tree


def traverse_tree_to_update_dist(
    node,
    matrix,
    method="percentile",
    percentile=50,
    rename_leaf_map=lambda x: x.split("-")[0],
):
    leaf_names = node.get_leaf_names()
    if rename_leaf_map is not None:
        sel_idx = np.array([rename_leaf_map(x) for x in leaf_names]).astype(int)
    else:
        sel_idx = np.array(leaf_names).astype(int)

    if len(sel_idx) > 1:
        data = pd.DataFrame(matrix[sel_idx][:, sel_idx].flatten()).dropna()[0].values
        if method == "percentile":
            node.dist = np.percentile(data, percentile)
        elif method == "min":
            node.dist = np.min(data)
        elif method == "mean":
            node.dist = np.mean(data)
        elif method == "median":
            node.dist = np.median(data)
        elif method == "max":
            node.dist = np.max(data)
        else:
            raise ValueError(
                "method should be among {percentile, mean, median, min, max}"
            )
    else:
        node.dist = np.nan
    for child in node.children:
        traverse_tree_to_update_dist(
            child, matrix, method=method, percentile=percentile
        )



import warnings

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from . import QC, metadata
from .settings import *

warnings.filterwarnings("ignore")
pd.options.mode.chained_assignment = None  # disable the SettingWithCopyWarning

####################################################

## This is meant for acc correction. No longer used

#####################################################


def order_by_FACS_sorting(df_sample):
    """
    Invert on the odd row, as in FACS
    """
    count = 0
    my_list = []
    for j, y in enumerate(["A", "B", "C", "D", "E", "F", "G", "H"]):
        temp = []
        for x in range(1, 13):
            temp.append(f"{y}{x}")
        if j % 2 == 0:
            my_list += temp
        else:
            my_list += list(np.array(temp)[::-1])
    df_tmp = pd.DataFrame({"FACS_order": my_list})
    df_sample["FACS_order"] = df_sample["sample"].apply(lambda x: x.split("_")[-1])
    return df_tmp.merge(df_sample, on="FACS_order")


def transform_rate(x, t1_over_t0, scale=1):
    """
    x is the input rate
    """
    offset = 10 ** (-10)
    if scale == 100:
        x = x / 100
        y = 1 - np.exp(-t1_over_t0 * np.log((1 + offset) / (1 + offset - x)))
        y = 100 * y
    else:
        y = 1 - np.exp(-t1_over_t0 * np.log((1 + offset) / (1 + offset - x)))

    return y


def estimate_t1_over_t0(global_rate_t0, global_rate_t1, scale=1):
    if scale == 100:
        global_rate_t0 = global_rate_t0 / 100
        global_rate_t1 = global_rate_t1 / 100

    t1_over_t0 = np.log(1.0001 - global_rate_t1) / np.log(1.0001 - global_rate_t0)
    return t1_over_t0


def QC_check_batch_correction(df_data, sample_1, sample_2, data_des=""):
    fig, axs = plt.subplots(1, 2, figsize=(10, 4))
    df_1 = df_data[df_data["sample"].isin([sample_1])]
    df_2 = df_data[df_data["sample"].isin([sample_2])]
    ax = sns.histplot(
        data=df_1, x="rate", label=sample_1, color="r", alpha=0.5, ax=axs[0]
    )
    ax = sns.histplot(
        data=df_2, x="rate", label=sample_2, color="b", alpha=0.5, ax=axs[0]
    )
    ax.legend()
    ax.set_title(f"Before correction: {data_des}")

    df_1 = df_data[df_data["sample"].isin([sample_1])]
    df_2 = df_data[df_data["sample"].isin([sample_2])]
    t1_over_t0 = estimate_t1_over_t0(
        df_2["rate"].mean(), df_1["rate"].mean(), scale=100
    )
    df_2["rate"] = df_2["rate"].apply(
        lambda x: transform_rate(x, t1_over_t0, scale=100)
    )
    sns.histplot(data=df_1, x="rate", label=sample_1, color="r", alpha=0.5, ax=axs[1])
    ax = sns.histplot(
        data=df_2, x="rate", label=sample_2, color="b", alpha=0.5, ax=axs[1]
    )
    ax.set_title(f"After correction: {data_des}")


def batch_correction_context_specific(df_input, df_sample, force_run=False):
    if (
        ("rate_before_correction" in df_input.columns)
        and ("t1_over_t0" in df_input.columns)
    ) and (not force_run):
        print("It seems that transformation has already been done. Abort")

    else:
        df_data = df_input.copy()
        df_data["rate_before_correction"] = df_data["rate"]
        df_data = df_data.merge(
            df_sample.filter(["sample", "lineage", "HQ"]), on="sample"
        )
        df_data = df_data[df_data["HQ"]]
        # obtain site-specific t1_over_t0 parameter
        df_tmp = (
            df_data.groupby(["sample", "anno", "lineage"])
            .agg(global_rate=("rate", "mean"))
            .reset_index()
        )
        df_tmp_2 = (
            df_data.groupby(["lineage", "anno"])
            .agg(lineage_rate=("rate", "mean"))
            .reset_index()
        )
        df_tmp = df_tmp.merge(df_tmp_2, on=["lineage", "anno"])
        anno_list = list(set(df_data["anno"]))
        df_tmp_list = []
        for anno in anno_list:
            df_anno = df_tmp[df_tmp["anno"] == anno]
            df_anno["t1_over_t0"] = estimate_t1_over_t0(
                df_anno["global_rate"], df_anno["lineage_rate"], scale=100
            )
            df_tmp_list.append(df_anno)
        df_tmp_new = pd.concat(df_tmp_list)
        df_data = df_data.merge(
            df_tmp_new.filter(
                ["sample", "anno", "global_rate", "lineage_rate", "t1_over_t0"]
            ),
            on=["sample", "anno"],
        )

        # perform transformation
        df_data["rate"] = transform_rate(
            df_data["rate"], df_data["t1_over_t0"], scale=100
        )

    return df_data


def batch_correction_global_time(df_input, df_sample, force_run=False, source="acc"):
    if ("rate_before_correction" in df_input.columns) and (not force_run):
        print("It seems that transformation has already been done. Abort")

    else:
        df_data = df_input.copy()
        df_data["rate_before_correction"] = df_data["rate"]
        df_sample_tmp = df_sample.copy()
        df_sample_tmp["rate_ref"] = 100 * df_sample_tmp[f"{source}_rate"]

        # obtain site-specific t1_over_t0 parameter
        df_tmp_2 = (
            df_sample_tmp[df_sample_tmp["HQ"]]
            .groupby(["lineage"])
            .agg(lineage_rate=("rate_ref", "mean"))
            .reset_index()
        )
        df_tmp = df_sample_tmp.filter(["lineage", "rate_ref", "sample", "HQ"]).merge(
            df_tmp_2, on=["lineage"]
        )
        # df_tmp["lineage_rate"] = 40
        # print("fix linenage rate to be constant")

        df_tmp["t1_over_t0"] = estimate_t1_over_t0(
            df_tmp["rate_ref"], df_tmp["lineage_rate"], scale=100
        )
        df_data = df_data.merge(df_tmp, on=["sample"])

        # perform transformation
        df_data["rate"] = transform_rate(
            df_data["rate"], df_data["t1_over_t0"], scale=100
        )
        df_data = df_data[df_data["HQ"]].drop("HQ", axis=1)
    return df_data


def log_transform(x, scale=100):
    if scale == 100:
        return np.log2(((x / 100) + 0.01) / (1 - (x / 100) + 0.01))
    else:
        return np.log2(((x) + 0.01) / (1 - (x) + 0.01))


def compare_global_rate_after_correction(df_data, N_thresh=1):
    df_data_tmp = df_data[df_data["N"] > N_thresh]
    fig, axs = plt.subplots(1, 2, figsize=(10, 4))
    ax = sns.boxplot(
        data=df_data_tmp.groupby(["sample", "lineage"])
        .agg(rate=("rate_before_correction", "mean"))
        .reset_index(),
        x="lineage",
        y="rate",
        ax=axs[0],
    )
    ax.figure.autofmt_xdate(rotation=90)
    ax.set_ylabel("Global acc rate (before)")
    ax.set_xlabel("")

    ax = sns.boxplot(
        data=df_data_tmp.groupby(["sample", "lineage"])
        .agg(rate=("rate", "mean"))
        .reset_index(),
        x="lineage",
        y="rate",
        ax=axs[1],
    )
    ax.figure.autofmt_xdate(rotation=90)
    ax.set_ylabel("Global acc rate (after)")
    ax.set_xlabel("")
    title = "_".join(list(set(df_data_tmp["anno"]))) + f"; N_thresh={N_thresh}"
    fig.suptitle(title)


def compare_element_rate_after_correction(
    df_data, N_thresh=50, SampleList=None, lineage_rate_key="rate_ref"
):
    df_data_tmp = df_data[df_data["N"] > N_thresh]
    flag = 0
    if SampleList is None:
        flag = 1
        df_tmp = (
            df_data_tmp.groupby(["sample", "lineage"])
            .agg({lineage_rate_key: "mean"})
            .reset_index()
        )
        sel_lineage = (
            df_tmp.groupby("lineage")
            .agg(mean=(lineage_rate_key, "mean"), std=(lineage_rate_key, "std"))["std"]
            .idxmax()
        )
        df_tmp = (
            df_tmp[df_tmp["lineage"] == sel_lineage]
            .set_index("sample")
            .sort_values(lineage_rate_key)
            .reset_index()
        )

        max_sp = df_tmp["sample"].iloc[0]
        min_sp = df_tmp["sample"].iloc[-1]
        mean_sp = df_tmp["sample"].iloc[int(len(df_tmp) / 2)]
        SampleList = [max_sp, min_sp, mean_sp]

    fig, axs = plt.subplots(1, 2, figsize=(10, 4))
    sns.histplot(
        data=df_data_tmp[df_data_tmp["sample"].isin(SampleList)],
        x="rate_before_correction",
        hue="sample",
        ax=axs[0],
    )
    sns.histplot(
        data=df_data_tmp[df_data_tmp["sample"].isin(SampleList)],
        x="rate",
        hue="sample",
        ax=axs[1],
    )
    title = "_".join(list(set(df_data_tmp["anno"]))) + f"; N_thresh={N_thresh}"
    if flag == 1:
        fig.suptitle(f"{sel_lineage}; " + title)
    else:
        fig.suptitle(title)
    plt.tight_layout()


##############################################

## Correlation metrics

##############################################


def compute_cell_cell_correlation(adata, lineage1, lineage2, nCG_thresh=10**6):
    from scipy.stats import pearsonr

    condition_1 = (adata.obs["lineage"] == lineage1) & (adata.obs["nGC"] > nCG_thresh)
    condition_2 = (adata.obs["lineage"] == lineage2) & (adata.obs["nGC"] > nCG_thresh)

    adata_tmp_1 = adata[condition_1]
    adata_tmp_2 = adata[condition_2]
    corr_list = []
    p_value_list = []
    cell_id_list = []
    for x in range(adata_tmp_1.shape[0]):
        X1 = adata_tmp_1.X[x, :].toarray()
        cell_id_1 = adata_tmp_1.obs_names[x]
        for y in range(1, adata_tmp_2.shape[0]):
            X2 = adata_tmp_2.X[y, :].toarray()
            cell_id_2 = adata_tmp_2.obs_names[y]
            sp_idx = (X1 > 0) & (X2 > 0)
            corr, pvalue = pearsonr(X1, X2)
            corr_list.append(corr)
            p_value_list.append(pvalue)
            cell_id_list.append((cell_id_1, cell_id_2))

    p_value_list = np.array(p_value_list)
    df_corr = pd.DataFrame(
        {"corr": corr_list, "pvalue": p_value_list, "sample": cell_id_list}
    )
    df_corr["logPvalue"] = -np.log(df_corr["pvalue"] + 10 ** (-10))
    df_corr = df_corr[df_corr["corr"] < 0.99]
    return df_corr


def compute_all_cell_cell_corr(adata, nCG_thresh=2 * 10**6):
    lineage_list = list(set(adata.obs["lineage"]))
    df_list = []
    pair_data = []
    count = 0
    for k in range(len(lineage_list)):
        count = count + 1
        lg1 = lineage_list[k]
        lg2 = lineage_list[k]
        df_corr = compute_cell_cell_correlation(adata, lg1, lg2, nCG_thresh=nCG_thresh)
        # pair_data.append((lg1,lg2))
        df_corr["lineage_pair"] = f"{lg1},{lg2}"
        df_corr["rank"] = count
        df_list.append(df_corr)

    for k in range(len(lineage_list)):
        for j in range(k, len(lineage_list)):
            lg1 = lineage_list[k]
            lg2 = lineage_list[j]
            df_corr = compute_cell_cell_correlation(
                adata, lg1, lg2, nCG_thresh=nCG_thresh
            )
            # pair_data.append((lg1,lg2))
            if lg1 != lg2:
                count = count + 1
                df_corr["lineage_pair"] = f"{lg1},{lg2}"
                df_list.append(df_corr)
                df_corr["rank"] = count

    df_final = pd.concat(df_list)

    fig, ax = plt.subplots(figsize=(12, 4))
    sns.boxplot(data=df_final, x="lineage_pair", y="corr", ax=ax)
    plt.xlabel("")
    plt.xticks(rotation=90)
    plt.ylabel("Cell-cell corr across DNA elements")

    return df_final


def compare_plate_effect_after_batch_correction(df, df_sample, transform=False):
    """
    df should be output from batch correction method:
    batch_correction_global_time
    """
    df = df.merge(
        df_sample.filter(["id_acc", "id_met", "id_rna", "lineage", "sample"]),
        on="sample",
    )

    if transform:
        df["m_orig"] = log_transform(df["rate_before_correction"])
        df["m"] = log_transform(df["rate"])

        df_plot = (
            df.groupby(["sample", "lineage"])
            .agg(mean_rate=("m_orig", "mean"))
            .reset_index()
            .sort_values("sample")
            .reset_index()
        )
        g = sns.relplot(
            kind="scatter", data=df_plot, x="index", y="mean_rate", hue="lineage"
        )
        g.ax.set_xlabel("Ordered sample id")
        g.ax.set_ylabel("Original transformed rate")

        df_plot = (
            df.groupby(["sample", "lineage"])
            .agg(mean_rate=("m", "mean"))
            .reset_index()
            .sort_values("sample")
            .reset_index()
        )
        g = sns.relplot(
            kind="scatter", data=df_plot, x="index", y="mean_rate", hue="lineage"
        )
        g.ax.set_xlabel("Ordered sample id")
        g.ax.set_ylabel("Corrected transformed rate")
    else:
        df_plot = (
            df.groupby(["sample", "lineage"])
            .agg(mean_rate=("rate_before_correction", "mean"))
            .reset_index()
            .sort_values("sample")
            .reset_index()
        )
        g = sns.relplot(
            kind="scatter", data=df_plot, x="index", y="mean_rate", hue="lineage"
        )
        g.ax.set_xlabel("Ordered sample id")
        g.ax.set_ylabel("Original rate")

        df_plot = (
            df.groupby(["sample", "lineage"])
            .agg(mean_rate=("rate", "mean"))
            .reset_index()
            .sort_values("sample")
            .reset_index()
        )
        g = sns.relplot(
            kind="scatter", data=df_plot, x="index", y="mean_rate", hue="lineage"
        )
        g.ax.set_xlabel("Ordered sample id")
        g.ax.set_ylabel("Corrected rate")


def generate_count_matrix(
    root_dir, anno, source="acc", batch_correction="lineage", min_obs_N=3
):
    df_sample = metadata.load_sample_info(root_dir)
    if source == "acc":
        print("load acc")
        df_data = metadata.load_acc(root_dir, anno)
    else:
        print("load met")
        df_data = metadata.load_met(root_dir, anno)

    plt.hist(np.log10(1 + df_data["N"]), bins=100)
    plt.title("N counts histogram for raw data")

    # select high-quality cells, and genomic locus with high coverage
    df_sample["HQ"] = df_sample["HQ"] & (df_sample["acc_rate"] > 0.1)
    df_data = df_data[
        (df_data[f"N"] >= min_obs_N)
        & df_data["sample"].isin(df_sample[df_sample["HQ"]]["sample"])
    ]  # & (df_data['lineage']=='HSC') ]

    if batch_correction == "lineage":
        print("lineage specific batch correction")
        df_data = QC.batch_correction_context_specific(
            df_data, df_sample, force_run=False
        )
        QC.compare_global_rate_after_correction(df_data, N_thresh=1)
    elif batch_correction == "global":
        print("global batch correction")
        df_data = QC.batch_correction_global_time(df_data, df_sample, source=source)
        QC.compare_global_rate_after_correction(df_data, N_thresh=1)
    else:
        print("no batch correction")

    df_counts_raw = df_data.pivot(index=["sample"], columns="id", values=f"rate")
    return df_counts_raw



import glob
import os

import multiprocess
import pandas as pd
from tqdm import tqdm

from . import metadata
from .settings import *


def generate_cell_by_region_matrix(
    data_path, selected_sample_list, annotation, threads=64
):

    pool = multiprocess.Pool(threads)

    ## annotate each individual pseudo-cell
    def annotate_single_sample(sample_tmp):
        command = f"Rscript {source_script_dir}/Rscripts/annotate/scNMT_annotate_individual_sample.R  --context CG  --basedir {data_path}  --featuresdir {default_feature_dir}  --sample {sample_tmp}  --annos {annotation}"
        os.system(command)

    pool.map(annotate_single_sample, selected_sample_list)

    ## Merge all annotations
    command = f"Rscript {source_script_dir}/Rscripts/annotate/scNMT_annotate_merge.R  --context CG  --basedir {data_path}  --featuresdir {default_feature_dir}   --annos {annotation}"
    os.system(command)
    print("-----JOB Finished-----")


def generate_TSS_profile(data_path, selected_sample_list, annotation, threads=64):

    data_output = f"{data_path}/results/metacc/profiles/data"
    sample_path = f"{data_path}/sample_sheet.tsv.gz"
    df_sample = metadata.initialize_sample_info(data_path)
    df_sample.to_csv(sample_path, sep="\t", index=False, compression="gzip")

    os.makedirs(data_output, exist_ok=True)

    def profile_function_tmp(sample):
        command = f"cd {source_script_dir}/Rscripts; Rscript profiles/calculate_only_met_profiles_SW.R --met_tile 100 --window_size 2000 --metadata {sample_path} \
        --anno {annotation} --out_dir {data_output} --featuredir {default_feature_dir}  --basedir {data_path} --sample {sample} "
        os.system(command)

    pool = multiprocess.Pool(threads)
    pool.map(profile_function_tmp, selected_sample_list)
    print("-----JOB Finished-----")


def generate_script_bulk_bigwig(
    root_path,
    source="acc",
    min_cells=5,
    group_by="lineage",
    sample_list=None,
    metadata_path_0=None,
    reference="mm10",
    step_size=None,
    scenario="all",
):
    print(f"script scenario: {scenario}")
    if scenario not in ["all", "only_bulk", "only_bigwig"]:
        raise ValueError("scenario must be in [all,only_bulk,only_bigwig]")

    os.makedirs(f"{root_path}/scripts", exist_ok=True)
    os.makedirs(f"{root_path}/acc/gpc_level/pseudobulk/{group_by}", exist_ok=True)
    os.makedirs(f"{root_path}/met/cpg_level/pseudobulk/{group_by}", exist_ok=True)

    if metadata_path_0 is None:
        metadata_path_0 = f"{root_path}/sample_sheet.tsv.gz"

    df_all_samples = pd.read_csv(metadata_path_0, sep="\t", compression="gzip")
    df_all_samples = df_all_samples[df_all_samples["HQ"]]
    if sample_list is None:
        metadata_path = f"{root_path}/sample_sheet.tsv.gz"
        sample_list = list(df_all_samples[group_by].unique())
        SAMPLE_LIST = "*".join(sorted(sample_list))
    else:
        SAMPLE_LIST = "*".join(sorted(sample_list))
        metadata_path = (
            f"{root_path}/scripts/sample_sheet_{group_by}_{SAMPLE_LIST}.tsv.gz"
        )
        df_all_samples[df_all_samples[group_by].isin(sample_list)].to_csv(
            metadata_path, sep="\t", compression="gzip", index=0
        )

    script_name = (
        f"{root_path}/scripts/generate_bulk_bigwig_{source}_{group_by}_{SAMPLE_LIST}.py"
    )

    if source == "met":
        if step_size is None:
            step_size = 500
    elif source == "acc":
        if step_size is None:
            step_size = 200
    else:
        raise ValueError("source must be acc or met")

    if scenario == "all":
        script = f"""
from methyltree  import help_functions as hf
data_path='{root_path}'
print(data_path)
hf.generate_bulk_data(data_path,source='{source}',min_cells='{min_cells}',group_by='{group_by}',metadata_path='{metadata_path}')
hf.generate_bigwig(data_path,source='{source}',group_by='{group_by}',sample_list='{SAMPLE_LIST}',step_size={step_size},reference='{reference}')
        """

    if scenario == "only_bulk":
        script = f"""
from methyltree  import help_functions as hf
data_path='{root_path}'
print(data_path)
hf.generate_bulk_data(data_path,source='{source}',min_cells='{min_cells}',group_by='{group_by}',metadata_path='{metadata_path}')
        """

    if scenario == "only_bigwig":
        script = f"""
from methyltree  import help_functions as hf
data_path='{root_path}'
print(data_path)
hf.generate_bigwig(data_path,source='{source}',group_by='{group_by}',sample_list='{SAMPLE_LIST}',step_size={step_size},reference='{reference}')
        """

    with open(script_name, "w") as w:
        w.write(script)

    return script_name


def generate_bigwig_core(
    input_dir, outdir, root_dir, source="met", sample_list=None, reference="mm10"
):
    if source == "met":
        step_size = 500
        min_rate_bigwig = 10
    elif source == "acc":
        step_size = 200
        min_rate_bigwig = 5
    else:
        raise ValueError("source must be acc or met")

    genome_size_ref = f"{source_script_dir}/reference/{reference}_genome.size"

    if sample_list is None:
        all_paths = sorted(glob.glob(f"{input_dir}/*.tsv.gz"))
        sample_list = [x.split("/")[-1].split(".tsv.gz")[0] for x in all_paths]

    print("sample list:", sample_list)
    for sample in sample_list:
        print(f"current sample: {sample}")
        command = f"cd {source_script_dir}/Rscripts; Rscript pseudobulk/create_bigwig_pseudobulk_SW.R --indir {input_dir} --outdir {outdir} \
        --bedGraphToBigWig /soft/bio/bedGraphToBigWig-2.8-bbi4/bedGraphToBigWig  --step_size {step_size}  --min_rate_bigwig {min_rate_bigwig} \
         --basedir {root_dir}  --samples {sample}  --genome_size_ref {genome_size_ref}"
        os.system(command)


def generate_bigwig(
    root_dir,
    source="acc",
    group_by="lineage",
    sample_list=None,
    step_size=None,
    reference="mm10",
):
    """
    sample_list: a string of samples joined by '*', like 'A*B*C'
    """

    if source == "met":
        input_dir = f"{root_dir}/met/cpg_level/pseudobulk/{group_by}"
        outdir = f"{root_dir}/met/bigwig"
        if step_size is None:
            step_size = 500
        min_rate_bigwig = 10
    elif source == "acc":
        input_dir = f"{root_dir}/acc/gpc_level/pseudobulk/{group_by}"
        outdir = f"{root_dir}/acc/bigwig"
        if step_size is None:
            step_size = 200
        min_rate_bigwig = 5
    else:
        raise ValueError("source must be acc or met")

    if sample_list is None:
        all_paths = sorted(glob.glob(f"{input_dir}/*.tsv.gz"))
        sample_list = [x.split("/")[-1].split(".tsv.gz")[0] for x in all_paths]
    else:
        sample_list = sample_list.split("*")

    genome_size_ref = f"{source_script_dir}/reference/{reference}_genome.size"
    print(f"genome ref path: {genome_size_ref}")

    print("sample list:", sample_list)
    for sample in sample_list:
        print(f"current sample: {sample}")
        command = f"cd {source_script_dir}/Rscripts; Rscript pseudobulk/create_bigwig_pseudobulk_SW.R --indir {input_dir} --outdir {outdir} \
        --bedGraphToBigWig /soft/bio/bedGraphToBigWig-2.8-bbi4/bedGraphToBigWig  --step_size {step_size}  --min_rate_bigwig {min_rate_bigwig} \
         --basedir {root_dir}  --samples {sample} --genome_size_ref {genome_size_ref}"
        os.system(command)


def generate_differential_analysis_script(
    annotation_list, root_dir, metadata_path=None
):
    if metadata_path is None:
        metadata_path = f"{root_dir}/sample_sheet.tsv.gz"

    script_list = []
    for anno_tmp in annotation_list:
        script_name = f"{root_dir}/scripts/diff_analysis_{anno_tmp}.py"
        script = f"""
from methyltree  import help_functions as hf
import pandas as pd
data_path='{root_dir}'
print(data_path)
df_sample=pd.read_csv('{metadata_path}',sep='\t',compression='gzip')
selected_fates=list(set(df_sample['lineage']))
selected_fates.remove('nan0')
print(selected_fates)
hf.differential_analysis(data_path,selected_fates,anno='{anno_tmp}',source='acc',min_cell_N=5,min_C_count=5,metadata_path='{metadata_path}')
hf.differential_analysis(data_path,selected_fates,anno='{anno_tmp}',source='met',min_cell_N=5,min_C_count=3,metadata_path='{metadata_path}')
        """
        with open(script_name, "w") as w:
            w.write(script)
        script_list.append(script_name)
    return script_list


def run_sbatch(
    command,
    sbatch_mode="intel-sc3",
    mem="10G",
    cores=2,
    time="01:0:0",
    job_name="sbatch",
):
    os.system("mkdir -p log")
    sbatch_command = f'sbatch -p {sbatch_mode} -c {cores} -t {time} --mem={mem} --job-name {job_name} --output=log/{job_name}-%j.o  --error=log/{job_name}-%j.e --mail-type=TIME_LIMIT_90,FAIL,END --wrap="{command}"'
    print(f"submit job:   {sbatch_command}")
    os.system(sbatch_command)


def merge_fastq(data_path_1, data_path_2, data_path_out, SampleList):
    """
    Merge fastq files from different sequencing run
    ```python
    import yaml
    import os
    root_dir='/storage/wangshouwenLab/wangshouwen/DATA/multiomics/20211027_scLimeCat_LARRY'
    with open(f"{root_dir}/result_DNA_merged/config.yaml", "r") as stream:
        file = yaml.safe_load(stream)
        SampleList_DNA = file["SampleList"]

    data_path_1=f'{root_dir}/raw_fastq'
    data_path_2=f'{root_dir}/reseq_data/raw_fastq'
    data_path_out=f'{root_dir}/merge/raw_fastq'

    merge_fastq(data_path_1,data_path_2,data_path_out,SampleList)
    ```
    """

    for sample in tqdm(SampleList):
        os.makedirs(f"{data_path_out}/{sample}", exist_ok=True)
        # os.system(f'rm {data_path}/{sample}_R1_001.fastq.gz')
        for source in ["R1", "R2"]:
            file_1 = f"{data_path_1}/{sample}/{sample}_{source}.fastq.gz"
            file_2 = f"{data_path_2}/{sample}/{sample}_{source}.fastq.gz"
            file_3 = f"{data_path_out}/{sample}/{sample}_{source}.fastq.gz"
            if os.path.exists(file_1) & os.path.exists(file_2):
                command = f"cat {file_1} {file_2}   > {file_3}"
                os.system(command)
                # hf.run_sbatch(command,mem='15G',cores=1,time='00:15:00',job_name=f'cat_{sample}_{source}')
            elif os.path.exists(file_1) and (not os.path.exists(file_2)):
                print(f"{sample} not exist in data_path_2. Direct copy")
                os.system(f"cp {file_1} {file_3}")
            elif os.path.exists(file_2) and (not os.path.exists(file_1)):
                print(f"{sample} not exist in data_path_1. Direct copy")
                os.system(f"cp {file_2} {file_3}")
            else:
                print(f"{sample} not exist in either data_path_1 or data_path_2")


def generate_bulk_data(
    root_dir,
    source="acc",
    min_cells=2,  # 5
    ncores=1,
    group_by="lineage",
    metadata_path=None,
):
    # log_dir=f'{root_dir}/log'
    # os.makedirs(log_dir,exist_ok=True)

    # metadata=f"{root_dir}/results/met/qc/sample_metadata_after_met_qc.txt.gz"
    if metadata_path is None:
        metadata_path = f"{root_dir}/sample_sheet.tsv.gz"
    if source == "met":
        input_dir = f"{root_dir}/met/cpg_level"
        outdir = f"{root_dir}/met/cpg_level/pseudobulk/{group_by}"
        context = "CG"
    elif source == "acc":
        input_dir = f"{root_dir}/acc/gpc_level"
        outdir = f"{root_dir}/acc/gpc_level/pseudobulk/{group_by}"
        context = "GC"
    else:
        raise ValueError("source must be acc or met")

    command = f"cd {source_script_dir}/Rscripts; Rscript pseudobulk/pseudobulk_metacc_SW.R --indir {input_dir} --metadata {metadata_path} --context {context} --group_by {group_by} \
        --ncores {ncores} --min_cells {min_cells} --outdir {outdir} --basedir {root_dir}"
    os.system(command)


def differential_analysis(
    root_dir: str,
    selected_fates: list,
    anno: str,
    source="acc",
    min_cell_N=5,
    min_C_count=5,
    metadata_path=None,
):
    """
    root_dir:
        directory to downstream_R/all_data
    selected_fates:
        A list of selected fates to perform differential analysis
    anno:
        annotation
    min_cell_N:
        minimum number of cells per group
    min_C_count:
        minimum C counts per cell per feature
    """
    if metadata_path is None:
        metadata_path = f"{root_dir}/sample_sheet.tsv.gz"
    if source == "met":
        data_dir = f"{root_dir}/met/feature_level"
    elif source == "acc":
        data_dir = f"{root_dir}/acc/feature_level"
    else:
        raise ValueError("source must be acc or met")

    all_files = [x for x in os.listdir(data_dir) if x.endswith(".tsv.gz") and anno in x]
    if len(all_files) == 0:
        raise ValueError(f"The selected annotation {anno} has not been computed yet")

    sel_N = len(selected_fates)
    for j in range(sel_N):
        for k in range(j + 1, sel_N):
            lineage_1 = selected_fates[j]
            lineage_2 = selected_fates[k]

            file_name = "*".join(sorted([lineage_1, lineage_2]))
            test_method = "binomial"  # binomial or t_test
            outfile = f"{root_dir}/{source}/differential/{anno}*{file_name}.txt"
            os.makedirs(os.path.dirname(outfile), exist_ok=True)
            if source == "acc":
                script = (
                    f"{source_script_dir}/diff/differential_analysis_acc.R"
                )
            else:
                script = (
                    f"{source_script_dir}/diff/differential_analysis_met.R"
                )
            os.system(
                f"""Rscript {script}   --data_dir {data_dir}  --sample_file {metadata_path}  --anno {anno}  --lineage1 {lineage_1} --lineage2 {lineage_2}    --min_cells {min_cell_N}  --min_C_count {min_C_count}  --outfile {outfile}   --test_method {test_method}"""
            )



import os
import random
import time

import numpy as np
import scanpy as sc
import scipy.sparse as ssp
from tqdm import tqdm


def select_remaining_values(value):
    remaining_values = [0, 1]
    if value in remaining_values:
        remaining_values.remove(value)
    selected_value = random.choice(remaining_values)
    return selected_value


def mutation_sequence(sequence, mutation_rate):
    new_sequence = sequence.copy()
    selected_sites = np.nonzero(np.random.rand(len(new_sequence)) < mutation_rate)[0]
    for i_site in selected_sites:
        new_sequence[i_site] = select_remaining_values(new_sequence[i_site])
    return new_sequence


def sigma(x, diff_sigma):
    #    return 0.1*np.sin(0.1*x)+0.1
    return diff_sigma


def progression(x, dL):
    #    return 0.1*np.sin(0.1*x)+1
    return dL


def transition_prob(x, y_vec, diff_sigma, dL):
    temp = []
    for y in y_vec:
        temp.append(
            np.exp(
                -((x - y + progression(x, dL)) ** 2) / (2 * (sigma(x, diff_sigma) ** 2))
            )
        )

    norm_Prob = np.array(temp) / np.sum(temp)

    return norm_Prob


def simulate_replication_next_time_point(
    x_input_list, cell_name_list_input, mutation_rate=0.01, progeny_N=2
):
    x_next_list = []
    cell_name_list_new = []

    for j, sequence in enumerate(x_input_list):
        old_name = cell_name_list_input[j]

        for k in range(progeny_N):
            new_sequence = sequence.copy()
            CpG_size = len(sequence)

            # mutate the CpG sites
            new_sequence = mutation_sequence(new_sequence, mutation_rate)

            x_next_list.append(new_sequence)
            cell_name_list_new.append(f"{old_name}-{k}")

    return x_next_list, cell_name_list_new


def simulate_discrete_differentiation_next_time_point(
    x_input_list,
    cell_name_list_input,
    mutation_rate=0.01,
    fate_CpG_frac=0.5,
    differentiation_prob=0.3,
    progeny_n=2,
):
    x_next_list = []
    cell_name_list_new = []

    for j, X in enumerate(x_input_list):
        old_name = cell_name_list_input[j]
        sequence = X[0]
        fate = X[1]

        # replication
        for k in range(progeny_n):
            new_sequence = sequence.copy()
            CpG_size = sequence.shape[0]

            # update fate in a probabilistic way
            if fate != -1:
                new_fate = fate  # already differentiated, keep the same fate
            else:
                if np.random.rand(1) < differentiation_prob:
                    new_fate = int(
                        np.random.rand(1) > 0.5
                    )  # choose fate between 0 and 1
                else:
                    new_fate = -1  # keep the un-differentiated fate

            # effect of differentiation on the CpG state
            if fate != new_fate:
                end_CpG = int(fate_CpG_frac * CpG_size)
                new_sequence[:end_CpG] = new_fate  # B branch

            # mutate the CpG sites
            new_sequence = mutation_sequence(new_sequence, mutation_rate)

            x_next_list.append([new_sequence, new_fate])
            cell_name_list_new.append(f"{old_name}-{k}")

    return x_next_list, cell_name_list_new


def simulate_bifurcation_next_time_point(x_input_list, cell_name_list, parameter):
    rna_diff_length = parameter["rna_diff_length"]
    rna_dL = parameter["rna_dL"]
    diff_sigma = parameter["diff_sigma"]
    progeny_N = parameter["progeny_N"]
    lattice = parameter["lattice"]
    bifurcation = parameter["bifurcation"]
    mutation_rate = parameter["mutation_rate"]
    fate_CpG_frac = parameter["fate_CpG_frac"]

    x_next_list = []
    cell_name_list_new = []

    for j, X in enumerate(x_input_list):
        old_name = cell_name_list[j]
        x = X[0]
        sequence = X[1]
        fate = X[2]
        ## this method is much faster
        if x < rna_diff_length - (
            1.5 * progression(x, rna_dL) + 5 * sigma(x, diff_sigma)
        ):
            new_x_list = np.random.normal(
                x + progression(x, rna_dL), sigma(x, diff_sigma), progeny_N
            )

        else:
            ## this method is much slower
            new_x_list = []
            for k in range(progeny_N):
                cum_Prob = np.cumsum(transition_prob(x, lattice, diff_sigma, rna_dL))
                # pick the first event that is larger than the random variable
                new_id = np.nonzero(cum_Prob > np.random.rand())[0][0].astype(int)
                new_x_list.append(lattice[new_id])

        ## loop for each progeny
        for k, new_x in enumerate(new_x_list):
            # update fate
            if x >= bifurcation:  # do not change fate, already committed
                new_fate = fate
            elif (x < bifurcation) and (new_x < bifurcation):
                new_fate = -1
            else:
                new_fate = np.random.randint(2)

            # update the CpG methylation due to changes of cell type
            new_sequence = sequence.copy()
            CpG_size = len(sequence)

            diff_frac_0 = (x - bifurcation) / (rna_diff_length - bifurcation)
            diff_frac_new = (new_x - bifurcation) / (rna_diff_length - bifurcation)

            start_CpG = int(diff_frac_0 * CpG_size * fate_CpG_frac)
            end_CpG = int(diff_frac_new * CpG_size * fate_CpG_frac)

            # already differentiated, modify a new region to indicate differentiation progression
            if (fate == new_fate) and (new_fate != -1) and (end_CpG > start_CpG):
                new_sequence[start_CpG:end_CpG] = new_fate

            # differentiation and fate change
            if (fate != new_fate) and (diff_frac_new >= 0):
                new_sequence[:end_CpG] = new_fate  # B branch

            # mutate the CpG sites
            new_sequence = mutation_sequence(new_sequence, mutation_rate)

            x_next_list.append([new_x, new_sequence, new_fate])
            cell_name_list_new.append(f"{old_name}-{k}")

    return x_next_list, cell_name_list_new


def replication_model(
    CpG_size=10000,
    initial_random_rate=0.05,
    clone_N=1,
    mutation_rate=0.01,
    generation=5,
    sequence_initial_orig=None,
):
    print("Generate new data")
    met_X = []
    met_names = []
    clone_id = []
    # initialize baseline methylation profile, same for all clones
    if sequence_initial_orig is None:
        sequence_initial_orig = (np.random.rand(CpG_size) > 0.5).astype(int)
    for j in range(clone_N):
        sequence_initial = mutation_sequence(sequence_initial_orig, initial_random_rate)

        x_next_list = [sequence_initial]
        cell_name_list = [f"{j}"]
        for _ in range(generation):
            x_next_list, cell_name_list = simulate_replication_next_time_point(
                x_next_list, cell_name_list, mutation_rate=mutation_rate, progeny_N=2
            )
        met_X = met_X + x_next_list
        met_names = met_names + cell_name_list

        clone_id = clone_id + [f"clone_{j}" for _ in range(len(cell_name_list))]

    adata_met = sc.AnnData(np.array(met_X))
    adata_met.obs_names = met_names
    adata_met.obs["clone_id"] = clone_id
    return adata_met


def discrete_differentiation_model(
    CpG_size=10000,
    initial_random_rate=0.05,
    clone_N=1,
    mutation_rate=0.01,
    fate_CpG_frac=0.5,
    differentiation_prob=0.3,
    generation=5,
    sequence_initial_orig=None,
):
    met_X = []
    met_names = []
    clone_id = []
    fate_list = []
    fate_CpG_N = int(fate_CpG_frac * CpG_size)

    # initialize methylation profile of the baseline profile, same for all clones
    if sequence_initial_orig is None:
        sequence_initial_orig = (np.random.rand(CpG_size) > 0.5).astype(
            float
        )  # need to use 'float', otherwise 0.5 will be forced to 0 if we use int structure
    for i in range(fate_CpG_N):
        sequence_initial_orig[i] = i % 2
    fate = -1
    for j in range(clone_N):
        sequence_initial = mutation_sequence(sequence_initial_orig, initial_random_rate)

        x_next_list = [[sequence_initial, fate]]
        cell_name_list = [f"{j}"]
        for _ in range(generation):
            (
                x_next_list,
                cell_name_list,
            ) = simulate_discrete_differentiation_next_time_point(
                x_next_list,
                cell_name_list,
                mutation_rate=mutation_rate,
                differentiation_prob=differentiation_prob,
                progeny_n=2,
                fate_CpG_frac=fate_CpG_frac,
            )
        met_X = met_X + [X[0] for X in x_next_list]
        fate_list = fate_list + [X[1] for X in x_next_list]
        met_names = met_names + cell_name_list

        clone_id = clone_id + [f"clone_{j}" for _ in range(len(cell_name_list))]

    fate_map = {-1: "stem", 0: "diff_A", 1: "diff_B"}
    adata_met = sc.AnnData(np.array(met_X))
    adata_met.obs_names = met_names
    adata_met.obs["clone_id"] = clone_id
    adata_met.obs["fate"] = [fate_map[_] for _ in fate_list]
    return adata_met


def single_cell_expansion(
    CpG_size,
    mutation_rate=0.001,
    generation=7,
    pruning_size=128,
    sequence_initial_orig=None,
):
    met_X = []
    met_names = []
    if sequence_initial_orig is None:
        sequence_initial_orig = (np.random.rand(CpG_size) > 0.5).astype(int)
    sequence_initial = mutation_sequence(
        sequence_initial_orig, mutation_rate
    )  # start from a single cell

    ## the first generation
    x_next_list = [sequence_initial]
    cell_name_list = [f"P0*"]
    for _ in range(generation):
        print(f"current generation: {_}")
        print(f"current cell number: {len(x_next_list)}")
        if len(x_next_list) > pruning_size:
            id_array = np.arange(len(x_next_list))
            sel_id = np.random.choice(id_array, pruning_size)
            x_next_list = np.array(x_next_list)[sel_id]
            cell_name_list = np.array(cell_name_list)[sel_id]
        x_next_list, cell_name_list = simulate_replication_next_time_point(
            x_next_list, cell_name_list, mutation_rate=mutation_rate, progeny_N=2
        )

    met_X = met_X + list(np.array(x_next_list))  # [sel_id])
    met_names = met_names + list(np.array(cell_name_list))  # [sel_id])

    adata_met = sc.AnnData(np.array(met_X))
    adata_met.obs_names = met_names
    adata_met.obs["clone_id"] = ["-".join(x.split("-")[:-3]) for x in met_names]
    return adata_met


def bifurcation_model(
    progeny_N=2,
    generation=5,
    sample_prob=0.5,
    clone_N=50,
    rna_diff_length=10,
    diff_sigma=0.5,
    repeat_N=1,
    rna_dL=1,
    CpG_size=10000,
    initial_random_rate=0.05,
    mutation_rate=0.01,
    re_simulate=False,
    data_dir="simulated_data",
    fate_CpG_frac=0.5,
    mRNA_bifurcation_frac=0.5,
    sequence_initial_orig=None,
):
    """
    Jointly simulate methylation, RNA, and lineage for individual clones in a bifurcation model

    Parameters
    ----------
    progeny_N:
        Fold change of clone size after each generation. 2 means that it will double its clone size after one generation.
    generation:
        Initital sampling time point. Unit: cell cycle. By default
        t2=t1+1.
    sample_prob:
        Probability to sample cells at t1
    clone_N:
        Total number of clones to simulate
    rna_diff_length:
        Total length of the 1-d differentiation manifold
    diff_sigma:
        Differentiation noise
    rna_dL:
        Step size of differentiation for one generation
    re_simulate:
        Simulate new data (do not load pre-saved datasets)

    Returns
    -------
    adata:
        An adata object with clonal matrix, time info etc. Ready to be plugged into CoSpar.
    """

    os.makedirs(data_dir, exist_ok=True)
    file_name = f"{data_dir}/simulated_clonal_data_bifurcation_M{clone_N}_progeny{progeny_N}_L{rna_diff_length}_dL{rna_dL}_diffSigma{diff_sigma}_t1{generation}_p1{sample_prob}_simuV{repeat_N}"

    if os.path.exists(file_name + "_clonal_annot.npz") and (not re_simulate):
        print("Load existing data")
        clone_annot = ssp.load_npz(file_name + "_clonal_annot.npz")
        simu_data = np.load(file_name + "_others.npz", allow_pickle=True)
        rna_X = simu_data["rna_X"]
        met_X = simu_data["met_X"]
        fate_list = simu_data["fate_list"]
        time_info = simu_data["time_info"]
        clone_id = simu_data["clone_id"]
        cell_name_list = simu_data["cell_name_list"]
        bifurcation = mRNA_bifurcation_frac * rna_diff_length

    else:
        print("Generate new data")
        t = time.time()

        # initialize position (RNA) of the founding cell
        n = (
            clone_N * (progeny_N ** (generation + 1)) * 20
        )  # resolution of grid for sampling
        dx = rna_diff_length / n
        lattice = np.linspace(0, rna_diff_length, n)

        max_n = int((rna_diff_length - generation * rna_dL) / dx)
        t0_id = np.sort(
            np.random.choice(max_n, clone_N, replace=False)
        )  # random sample of indices
        x_t0_array = np.array(
            lattice[t0_id]
        )  # position of initial barcoding for each founding cell

        bifurcation = mRNA_bifurcation_frac * rna_diff_length

        parameter = {}
        parameter["rna_diff_length"] = rna_diff_length
        parameter["rna_dL"] = rna_dL
        parameter["diff_sigma"] = diff_sigma
        parameter["progeny_N"] = progeny_N
        parameter["dx"] = dx
        parameter["lattice"] = lattice
        parameter["bifurcation"] = bifurcation
        parameter["mutation_rate"] = mutation_rate
        parameter["fate_CpG_frac"] = fate_CpG_frac

        rna_X = []
        met_X = []
        fate_list = []
        time_info = []
        clone_id = []
        temp_clone_id = 0
        previouse_cell_N = 0
        cell_name_list = []

        # initialize methylation profile of the baseline profile, same for all clones
        if sequence_initial_orig is None:
            sequence_initial_orig = (np.random.rand(CpG_size) > 0.5).astype(
                float
            )  # need to use 'float', otherwise 0.5 will be forced to 0 if we use int structure
        fate_CpG_N = int(fate_CpG_frac * CpG_size)
        sequence_initial_orig[:fate_CpG_N] = 0.5

        ## simulate the multi-generational drift separately for each clone
        for k in tqdm(range(clone_N)):
            x_t0 = x_t0_array[k]
            # print("Current clone number:", m31)
            x_next = []
            cell_name_tmp = [f"{k}"]

            sequence_initial = mutation_sequence(
                sequence_initial_orig, initial_random_rate
            )

            if x_t0 < bifurcation:
                x_next.append([x_t0, sequence_initial, -1])  # fate=-1
            else:
                fate = np.random.randint(2)

                diff_frac = (x_t0 - bifurcation) / (rna_diff_length - bifurcation)
                end_CpG = int(diff_frac * fate_CpG_N)
                sequence_initial[:end_CpG] = fate

                x_next.append([x_t0, sequence_initial, fate])

            for j in range(generation):
                x_next, cell_name_tmp = simulate_bifurcation_next_time_point(
                    x_next, cell_name_tmp, parameter
                )

            sel_idx = np.random.rand(len(x_next)) < sample_prob
            sel_id = list(np.nonzero(sel_idx)[0].astype(int))

            for _ in sel_id:
                rna_X.append(x_next[_][0])
                met_X.append(x_next[_][1])
                fate_list.append(x_next[_][2])
                cell_name_list.append(cell_name_tmp[_])

            time_info = time_info + [1 for _ in range(len(sel_id))]

            clone_id = clone_id + [
                temp_clone_id for _ in range(len(time_info) - previouse_cell_N)
            ]
            previouse_cell_N = len(time_info)
            temp_clone_id = temp_clone_id + 1

        ### Generate clonal data matrix
        clone_annot = np.zeros((len(time_info), len(x_t0_array)))
        for j in range(len(time_info)):
            clone_annot[j, clone_id[j]] = 1

        clone_annot = ssp.csr_matrix(clone_annot)

        ssp.save_npz(file_name + "_clonal_annot.npz", clone_annot)
        np.savez(
            file_name + "_others.npz",
            rna_X=np.array(rna_X),
            met_X=np.array(met_X),
            fate_list=np.array(fate_list),
            time_info=np.array(time_info),
            clone_id=np.array(clone_id),
            cell_name_list=cell_name_list,
        )
        print("Time elapsed for generating clonal data: ", time.time() - t)

    ## transform the RNA data to 50-d
    UMAP_noise = 0.2  #
    rna_X = np.array(rna_X)
    state_info = np.array(fate_list)
    x_2d = np.zeros(len(rna_X))
    y_2d = np.zeros(len(rna_X))
    idx = state_info == -1
    x_2d[idx] = rna_X[idx] - bifurcation
    y_2d[idx] = 0 * rna_X[idx]
    idx = state_info == 0
    x_2d[idx] = 0.5 * (rna_X[idx] - bifurcation)
    y_2d[idx] = 0.5 * (rna_X[idx] - bifurcation)
    idx = state_info == 1
    x_2d[idx] = 0.5 * (rna_X[idx] - bifurcation)
    y_2d[idx] = -0.5 * (rna_X[idx] - bifurcation)

    X_high_dim = np.zeros((len(x_2d), 50))
    X_high_dim[:, 0] = x_2d
    X_high_dim[:, 1] = y_2d

    for j in range(48):
        X_high_dim[:, j + 2] = UMAP_noise * np.random.randn(len(x_2d))

    fate_map = {-1: "stem", 0: "diff_A", 1: "diff_B"}
    adata_rna = sc.AnnData(X_high_dim)
    adata_met = sc.AnnData(np.array(met_X))

    for adata_tmp in [adata_rna, adata_met]:
        adata_tmp.obsm["X_clone"] = clone_annot
        adata_tmp.obs["time_info"] = time_info
        adata_tmp.obs["fate"] = np.array([fate_map[_] for _ in state_info])
        adata_tmp.obsm["X_emb"] = X_high_dim[:, :2]
        adata_tmp.uns["data_des"] = ["bifurcation"]
        adata_tmp.obsm["X_rna_high_da"] = X_high_dim
        adata_tmp.obsm["X_rna_1d"] = rna_X
        adata_tmp.obs["clone_id"] = np.array([f"clone_{_}" for _ in clone_id])
        adata_tmp.obsm["X_met"] = np.array(met_X)
        adata_tmp.obs_names = cell_name_list
        adata_tmp.obs["lineage"] = cell_name_list

    return adata_rna, adata_met



import os

import numpy as np
import pandas as pd
import scipy.stats as stats
import seaborn as sns
from matplotlib import pyplot as plt
from scipy import sparse as sci_sparse
from tqdm import tqdm

from .metadata import *
from .settings import *

## First section is basic QC, and the second section is further QC like batch effect,
## checking various correlation metrics.


##############################################

## Basic QC, like mapping efficiency, CpG and GpC counts, MT acc rate vs global rate etc.

##############################################


def generate_chr_stats(compact_file_dir, recompute=False):
    if os.path.exists(f"{compact_file_dir}/df_chr_rate.csv") and (not recompute):
        df_chr_rate = pd.read_csv(f"{compact_file_dir}/df_chr_rate.csv", index_col=0)
        df_chr_nCG = pd.read_csv(f"{compact_file_dir}/df_chr_nCG.csv", index_col=0)
    else:
        file_list = [
            x.split(".npz")[0]
            for x in os.listdir(compact_file_dir)
            if x.endswith(".npz")
        ]  # and x.startswith('chr') and (x not in ['chrX.npz','chrY.npz','chrM.npz'])]
        rate_dict = {}
        nCG_dict = {}
        for _ in tqdm(file_list):
            mat = sci_sparse.load_npz(f"{compact_file_dir}/{_}.npz")
            # mat=mat[abs(mat).sum(1).A.flatten()>0]
            diff = mat.sum(0)
            tot_sites = abs(mat).sum(0)
            met_rate = (diff + tot_sites) / (2 * tot_sites)
            rate_dict[_] = list(met_rate.A[0])
            nCG_dict[_] = list(tot_sites.A[0])
        df_chr_rate = pd.DataFrame(rate_dict)
        df_chr_rate.to_csv(f"{compact_file_dir}/df_chr_rate.csv")
        df_chr_nCG = pd.DataFrame(nCG_dict)
        df_chr_nCG.to_csv(f"{compact_file_dir}/df_chr_nCG.csv")
    return df_chr_rate, df_chr_nCG


def extract_mapping_efficiency_core(file_path, mapping_mode="SE"):
    """
    Given the mapping_efficiency summary file, return a dataframe
    """

    ## convert to pandas file
    raw_data = (
        open(file_path).read().split("\n-------------------------------------")[1:]
    )

    sample_names = []
    mapping_efficiency_R1 = []
    mapping_efficiency_R2 = []
    total_C_R1 = []
    total_C_R2 = []

    for line in raw_data:
        temp = line.split("\n-------------")
        sample_names.append(temp[0].split("----------------------------")[0])

        mapping_efficiency_R1.append(
            temp[1].split("\nMapping efficiency:\t")[-1].split("%")[0]
        )
        total_C_R1.append(temp[1].split("Total number of C's analysed:\t")[-1])

        if mapping_mode == "SE":
            mapping_efficiency_R2.append(
                temp[2].split("\nMapping efficiency:\t")[-1].split("%")[0]
            )
            total_C_R2.append(temp[2].split("Total number of C's analysed:\t")[-1])

    if mapping_mode == "SE":
        df_info = pd.DataFrame(
            {
                "Mapping_R1": mapping_efficiency_R1,
                "Mapping_R2": mapping_efficiency_R2,
                "Total_C_R1": total_C_R1,
                "Total_C_R2": total_C_R2,
            }
        )
        df_info = df_info.replace("R2-unmapped--------", "nan")
        df_info = df_info.replace("R1-unmapped--------", "nan")
        df_info = df_info.dropna()
        df_info = df_info.astype(float)
        df_info["Mapping_all"] = df_info["Mapping_R1"] + df_info["Mapping_R2"]
        df_info["Total_C_all"] = df_info["Total_C_R1"] + df_info["Total_C_R2"]

    else:
        df_info = pd.DataFrame(
            {
                "Mapping_all": mapping_efficiency_R1,
                "Total_C_all": total_C_R1,
            }
        )
        df_info = df_info.replace("---PE-------------", "nan")
        df_info = df_info.dropna()
        df_info = df_info.astype(float)

    df_info["sample"] = sample_names
    df_info = df_info.loc[
        :,
        [
            "sample",
            "Mapping_all",
            "Total_C_all",
        ],
    ]

    return df_info


def extract_CpG_CHG_CHH_rate(file_path):
    """
    Given the rate file, return a dataframe
    """

    ## convert to pandas file
    raw_data = (
        open(file_path).read().split("\n-------------------------------------")[1:]
    )

    sample_names = []
    CpG_rate = []
    CHG_rate = []
    CHH_rate = []

    for line in raw_data:
        temp = line.split("-------------")
        sample_names.append(temp[0])

        CpG_rate.append(
            temp[-1].split("\nC methylated in CpG context:\t")[-1].split("%")[0]
        )
        CHG_rate.append(
            temp[-1].split("\nC methylated in CHG context:\t")[-1].split("%")[0]
        )
        CHH_rate.append(
            temp[-1].split("\nC methylated in CHH context:\t")[-1].split("%")[0]
        )

    df_info = pd.DataFrame(
        {
            "CpG_rate": CpG_rate,
            "CHG_rate": CHG_rate,
            "CHH_rate": CHH_rate,
        }
    )

    df_info = df_info.dropna()
    df_info = df_info.astype(float)
    df_info["sample"] = sample_names

    return df_info


def QC_mapping_efficiency(
    root_dir, force_run=False, mapping_mode="SE", genome_reference="mm10"
):
    """
    root_dir: directory leading to the folder of 'met,acc'
    """

    df_sample = load_sample_info(root_dir)
    SampleList = df_sample["sample"].to_numpy()
    file_path = f"{root_dir}/results/all_reports_mapping_efficiency.txt"
    output_file = f"{root_dir}/results/mapping_efficiency.csv"
    if os.path.exists(output_file):
        df_info = pd.read_csv(output_file)
        SampleList = SampleList[~np.in1d(SampleList, df_info["sample"])]
        if len(SampleList) > 0:
            print(f"There are {len(SampleList)} uncomputed samples")
            force_run = True

    if (not os.path.exists(output_file)) or force_run:
        print(f"Recompute!")
        os.system(f"mkdir -p {root_dir}/results")
        command = f"sh {source_script_dir}/DNA/extract_reports.sh {root_dir}/../../bismark  {root_dir}/results {mapping_mode}"
        os.system(command)
        df_info = extract_mapping_efficiency_core(file_path, mapping_mode=mapping_mode)
        df_info.to_csv(output_file, index=0)

    plot_key = "Mapping_all"
    df_info_temp = df_info.sort_values(plot_key).reset_index().dropna()
    ax = sns.scatterplot(x=df_info_temp.index, y=df_info_temp[plot_key], s=10)
    ax.set_xlabel("Sample ordering")
    ax.set_ylabel("Mapping efficiency")
    return df_info


def QC_CpG_CHG_CHH_rate(root_dir, mapping_mode="SE", force_run=False):
    """
    root_dir: directory leading to the folder of 'met,acc'
    """

    df_sample = load_sample_info(root_dir)
    SampleList = df_sample["sample"].to_numpy()
    file_path = f"{root_dir}/results/all_reports_methylation_rate.txt"
    output_file = f"{root_dir}/results/CpG_CHG_CHH_rate.csv"
    if os.path.exists(output_file):
        df_info = pd.read_csv(output_file)
        SampleList = SampleList[~np.in1d(SampleList, df_info["sample"])]
        if len(SampleList) > 0:
            print(f"There are {len(SampleList)} uncomputed samples")
            force_run = True

    if (not os.path.exists(output_file)) or force_run:
        print(f"Recompute!")
        command = f"sh {source_script_dir}/DNA/extract_reports.sh {root_dir}/../../bismark  {root_dir}/results  {mapping_mode}"
        # print(command)
        os.system(f"mkdir -p {root_dir}/results")
        os.system(command)

        df_info = extract_CpG_CHG_CHH_rate(file_path)
        df_info.to_csv(output_file, index=0)

    return df_info


def QC_coverage_report_core(
    root_dir,
    source="met",
    feature_dir=default_feature_dir,
    weight_cutoff=1,
    feature_list=[
        "prom_2000_2000",
        "genebody",
        "prom_2000_2000_cgi",
        "prom_2000_2000_noncgi",
        "CGI",
        "LINE",
        "LTR",
    ],
):
    df_list = []
    for feature in feature_list:
        print(f"---------Current feature: {feature} -----------")

        if source == "met":
            df_data = load_met(root_dir, feature)
        elif source == "acc":
            df_data = load_acc(
                root_dir, feature
            )  # ["sample", "id", "anno", "Nmet", "N", "rate"],
        else:
            raise ValueError

        df_temp = df_data[df_data["N"] >= weight_cutoff]
        df_temp_v2 = (
            df_temp.groupby(["sample", "anno"])
            .agg(
                unique_id_N=("id", "count"),
                average_reads=("N", "mean"),
                average_rate=("rate", "mean"),
            )
            .reset_index()
        )

        bed_info = pd.read_table(f"{feature_dir}/{feature}.bed")
        bed_info.columns = ["chr", "start", "end", "strand", "id", "region"]
        tot_anno_N = len(list(set(bed_info["id"])))
        df_temp_v2["coverage"] = df_temp_v2["unique_id_N"] / tot_anno_N
        df_list.append(df_temp_v2)

    df_merge = pd.concat(df_list, ignore_index=True)
    return df_merge


def QC_coverage_report(
    root_dir,
    source="met",
    feature_dir=default_feature_dir,
    weight_cutoff=1,
    feature_list=[
        "prom_2000_2000",
        "genebody",
        "prom_2000_2000_cgi",
        "prom_2000_2000_noncgi",
        "CGI",
        "LINE",
        "LTR",
    ],
    force_run=False,
):
    """
    root_dir should be where 'acc, met' folder is.
    """

    df_sample = load_sample_info(root_dir)
    SampleList = df_sample["sample"].to_numpy()
    print(f"In total {len(set(SampleList))} samples")

    file_path = (
        f"{root_dir}/results/feature_coverage_{source}_cutoff{weight_cutoff}.csv"
    )
    if os.path.exists(file_path) and (not force_run):
        df_merge = pd.read_csv(file_path)
        # extract the un-computed samples
        SampleList = SampleList[~np.in1d(SampleList, df_merge["sample"].unique())]
        if len(SampleList) > 0:
            print(f"Sample list has been updated! Compute!")
            df_merge = QC_coverage_report_core(
                root_dir,
                feature_dir=feature_dir,
                source=source,
                weight_cutoff=weight_cutoff,
                feature_list=feature_list,
            )

    else:
        df_merge = QC_coverage_report_core(
            root_dir,
            source=source,
            feature_dir=feature_dir,
            weight_cutoff=weight_cutoff,
            feature_list=feature_list,
        )

    df_merge.to_csv(file_path, index=0)
    ax = sns.violinplot(
        data=df_merge[df_merge["anno"].isin(feature_list)], x="anno", y="coverage"
    )
    plt.xticks(rotation=90)
    ax.set_ylabel(f"Coverage (>={weight_cutoff}C)")
    ax.set_xlabel("")
    ax.set_title(source)

    return df_merge


def QC_profile_core(root_dir, SampleList, annotation="prom_2000_2000"):
    """
    Compute methylation and accessibility profiles for a given sample list, and return
    the data in a dataframe.
    """

    # check samples
    un_ready_samples = []
    ready_samples = []

    corr_list = []
    site_N = []
    GC_min = []
    GC_max = []
    CG_min = []
    CG_max = []
    for j in tqdm(range(len(SampleList))):
        sample = SampleList[j]
        data_path = f"{root_dir}/results/metacc/profiles/data/precomputed_metacc_{annotation}_{sample}.txt.gz"
        if not os.path.exists(data_path):
            un_ready_samples.append(sample)
        else:
            df = pd.read_csv(data_path)
            df_out = (
                df.groupby(["cell", "anno", "dist", "context"])
                .agg(mean_rate=("rate", "mean"))
                .reset_index()
            )
            shared_sites = set(df_out[df_out["context"] == "CG"]["dist"]).intersection(
                set(df_out[df_out["context"] == "GC"]["dist"])
            )
            if len(shared_sites) > 5:  # avoid division error from correlation
                df_CG = df_out[df_out["context"] == "CG"]
                df_GC = df_out[df_out["context"] == "GC"]
                shared_site_N = len(shared_sites)
                df_CG_sub = df_CG[df_CG["dist"].isin(shared_sites)]
                df_GC_sub = df_GC[df_GC["dist"].isin(shared_sites)]
                corr_met = np.corrcoef(abs(df_CG_sub["dist"]), df_CG_sub["mean_rate"])[
                    0, 1
                ]
                corr_acc = np.corrcoef(abs(df_GC_sub["dist"]), df_GC_sub["mean_rate"])[
                    0, 1
                ]
                corr_met_acc = np.corrcoef(
                    df_CG_sub["mean_rate"], df_GC_sub["mean_rate"]
                )[0, 1]
                corr_list.append([corr_met, corr_acc, corr_met_acc])
                site_N.append(shared_site_N)
                GC_min.append(df_GC["mean_rate"].min())
                GC_max.append(df_GC["mean_rate"].max())
                CG_min.append(df_CG["mean_rate"].min())
                CG_max.append(df_CG["mean_rate"].max())
                ready_samples.append(sample)

    if len(un_ready_samples) > 0:
        print(f"These samples are not ready: {un_ready_samples}")

    if len(corr_list) > 0:
        corr_list = np.array(corr_list)
        df = pd.DataFrame(
            {
                "sample": ready_samples,
                "corr_met": corr_list[:, 0],
                "corr_acc": corr_list[:, 1],
                "corr_met_acc": corr_list[:, 2],
                "shared_sites": site_N,
                "acc_min": GC_min,
                "acc_max": GC_max,
                "met_min": CG_min,
                "met_max": CG_max,
            }
        )
    else:
        df = pd.DataFrame(
            None,
            columns=[
                "sample",
                "corr_met",
                "corr_acc",
                "corr_met_acc",
                "shared_sites",
                "acc_min",
                "acc_max",
                "met_min",
                "met_max",
            ],
        )

    return df


def QC_profile_core_only_met(root_dir, SampleList, annotation="prom_2000_2000"):
    corr_list = []
    ready_samples = []
    un_ready_samples = []
    CG_min = []
    CG_max = []
    for j in tqdm(range(len(SampleList))):
        sample = SampleList[j]
        profiles_data_path = f"{root_dir}/results/metacc/profiles/data/precomputed_metacc_{annotation}_{sample}.txt.gz"
        if not os.path.exists(profiles_data_path):
            un_ready_samples.append(sample)
        else:
            df_profile = pd.read_csv(profiles_data_path)
            df_out = (
                df_profile.groupby(["cell", "anno", "dist", "context"])
                .agg(mean_rate=("rate", "mean"))
                .reset_index()
            )
            df_CG = df_out[df_out["context"] == "CG"]
            if len(df_CG) > 5:
                corr_met = np.corrcoef(abs(df_CG["dist"]), df_CG["mean_rate"])[0, 1]
            else:
                corr_met = np.nan
            corr_list.append(corr_met)
            CG_min.append(df_CG["mean_rate"].min())
            CG_max.append(df_CG["mean_rate"].max())
            ready_samples.append(sample)

    if len(un_ready_samples) > 0:
        print(f"These samples are not ready: {un_ready_samples}")

    if len(corr_list) > 0:
        corr_list = np.array(corr_list)
        df = pd.DataFrame(
            {
                "sample": ready_samples,
                "corr_met": corr_list,
                "met_min": CG_min,
                "met_max": CG_max,
            }
        )
        df["met_range"] = df["met_max"] - df["met_min"]
    else:
        df = pd.DataFrame(
            None,
            columns=["sample", "corr_met", "met_min", "met_max", "met_range"],
        )
    return df


def QC_profile(root_dir, force_run=False, annotation="prom_2000_2000", only_met=False):
    df_sample = load_sample_info(root_dir)
    SampleList = df_sample["sample"].to_numpy()
    print(f"In total {len(set(SampleList))} samples")

    file_path = f"{root_dir}/results/profile_summary_{annotation}.csv"
    if os.path.exists(file_path) and (not force_run):
        df_0 = pd.read_csv(file_path)
        # extract the un-computed samples
        SampleList = SampleList[~np.in1d(SampleList, df_0["sample"])]
        if len(SampleList) > 0:
            print(f"Compute remaining {len(SampleList)} samples")
            print(SampleList)
            if only_met:
                df_1 = QC_profile_core_only_met(
                    root_dir, SampleList, annotation=annotation
                )
            else:
                df_1 = QC_profile_core(root_dir, SampleList, annotation=annotation)
            df = pd.concat([df_0, df_1], ignore_index=True)
        else:
            df = df_0
    else:
        if only_met:
            df = QC_profile_core_only_met(root_dir, SampleList, annotation=annotation)
        else:
            df = QC_profile_core(root_dir, SampleList, annotation=annotation)

    df.to_csv(file_path, index=0)
    return df


def plot_profile(root_dir, sample, annotation="prom_2000_2000"):
    data_path = f"{root_dir}/results/metacc/profiles/data/precomputed_metacc_{annotation}_{sample}.txt.gz"
    df = pd.read_csv(data_path)
    df_out = (
        df.groupby(["cell", "anno", "dist", "context"])
        .agg(mean_rate=("rate", "mean"))
        .reset_index()
    )
    sns.scatterplot(data=df_out, x="dist", y="mean_rate", hue="context")

    # .merge(df[df['context']=='GC'],on=['cell','id','anno','dist'])


def QC_count_reads(root_dir, force_run=False):
    """
    root_dir should be the directory to 'config.yaml'
    Running this is too slow. So, we will not use this.
    """

    file_path = f"{root_dir}/results/read_counts.csv"
    if os.path.exists(file_path) and (not force_run):
        df = pd.read_csv(file_path)
    else:
        df_sample = load_sample_info(root_dir)
        SampleList = df_sample["sample"].to_numpy()

        output_file = f"{root_dir}/read_counts.txt"
        os.system(f"echo '' > {output_file}")  # create an empty file

        for j in tqdm(range(len(SampleList))):
            sample = SampleList[j]
            file = f"{root_dir}/../raw_fastq/{sample}/{sample}_R1.fastq.gz"
            os.system(
                f"gzip -d -c {file} | wc -l >> {output_file}"
            )  # create an empty file

        # process the data
        df = pd.DataFrame(
            pd.read_csv(f"{root_dir}/read_counts.txt", sep="\t", header=None)[0]
            .apply(lambda x: x.split(" "))
            .to_list()
        ).rename(columns={0: "line_N", 1: "file_name"})
        df["read_N"] = df["line_N"].astype(int) / 4
        df["sample"] = df["file_name"].apply(
            lambda x: x.split("/")[-1].split("_R1.fastq.gz")[0]
        )
        df.to_csv(file_path, index=0)

        return df


def file_size_core(root_dir, SampleList):
    file_size_list = []
    file_name_list = []
    for j in tqdm(range(len(SampleList))):
        sample = SampleList[j]
        file = f"{root_dir}/../../../raw_fastq/{sample}/{sample}_R1.fastq.gz"
        file_size_list.append(os.path.getsize(file))
        file_name_list.append(f"{sample}_R1.fastq.gz")

    file_size_list = np.array(file_size_list)
    # process the data
    df = pd.DataFrame(
        {
            "sample": SampleList,
            "FileName": file_name_list,
            "file_size (byte)": file_size_list,
            "file_size (M)": np.round(file_size_list / 10**6),
            "file_size (G)": np.round(file_size_list / 10**9),
        }
    )
    return df


def QC_file_size(root_dir, force_run=False):
    """
    root_dir should be the directory to 'met,acc' folder
    """

    df_sample = load_sample_info(root_dir)
    SampleList = df_sample["sample"].to_numpy()
    print(f"In total {len(set(SampleList))} samples")

    file_path = f"{root_dir}/results/sample_size.csv"
    if os.path.exists(file_path) and (not force_run):
        df_0 = pd.read_csv(file_path)
        # extract the un-computed samples
        SampleList = SampleList[~np.in1d(SampleList, df_0["sample"])]
        if len(SampleList) > 0:
            print(f"Compute remaining {len(SampleList)} samples")
            df_1 = file_size_core(root_dir, SampleList)
            df = pd.concat([df_0, df_1], ignore_index=True)
        else:
            df = df_0
    else:
        df = file_size_core(root_dir, SampleList)

    df.to_csv(file_path, index=0)
    return df


def extract_count_statistics(root_dir, SampleList, source="met", cores=64):
    """
    Extract the CpG and GpC count number and global met/acc rate,
    as well as rate for MT and lambda DNA for a given SampleList.

    Do not save, but return a dataframe
    """

    def extract_count_statistics_core(sample):
        if source == "met":
            file_name_tmp = f"{root_dir}/met/cpg_level/{sample}.tsv.gz"
        elif source == "acc":
            file_name_tmp = f"{root_dir}/acc/gpc_level/{sample}.tsv.gz"
        else:
            raise ValueError("source should be {met,acc}")

        # print(file_name_tmp)
        df_tmp = pd.read_csv(
            file_name_tmp,
            sep="\t",
            compression="gzip",
            dtype={
                "chr": str,
                "pos": int,
                "met_reads": int,
                "nonmet_reads": int,
                "rate": float,
            },
        )
        unique_C = len(df_tmp)
        Nmet_reads = df_tmp["met_reads"].astype(int).sum()
        Nnonmet_reads = df_tmp["nonmet_reads"].astype(int).sum()
        N_tot_reads = Nmet_reads + Nnonmet_reads
        rates_tmp = df_tmp["rate"].mean()

        unique_chr = list(df_tmp["chr"].unique())
        if "lambda" in unique_chr:
            lambda_label = "lambda"
        elif "lambda_NEB" in unique_chr:
            lambda_label = "lambda_NEB"
        else:
            lambda_label = "J02459"

        if "chrM" in unique_chr:
            MT_label = "chrM"
        elif "MT" in unique_chr:
            MT_label = "MT"
        else:
            MT_label = "M"

        if "L09137" in unique_chr:
            puc19_label = "L09137"
        else:
            puc19_label = "pUC19"

        MT_rate_tmp = df_tmp[df_tmp["chr"] == MT_label]["rate"].mean()
        df_lambda = df_tmp[df_tmp["chr"] == lambda_label]
        lambda_rate_tmp = df_lambda["rate"].mean()
        lambda_CpG_count = len(df_lambda)
        lambda_read_count = (
            df_lambda["met_reads"].sum() + df_lambda["nonmet_reads"].sum()
        )

        df_puc19 = df_tmp[df_tmp["chr"] == puc19_label]
        puc19_rate_tmp = df_puc19["rate"].mean()
        puc19_CpG_count = len(df_puc19)
        puc19_read_count = df_puc19["met_reads"].sum() + df_puc19["nonmet_reads"].sum()
        return [
            source,
            sample,
            unique_C,
            Nmet_reads,
            Nnonmet_reads,
            N_tot_reads,
            rates_tmp,
            MT_rate_tmp,
            lambda_rate_tmp,
            lambda_CpG_count,
            lambda_read_count,
            puc19_rate_tmp,
            puc19_CpG_count,
            puc19_read_count,
        ]

    from joblib import Parallel, delayed

    data_list = Parallel(n_jobs=cores)(
        delayed(extract_count_statistics_core)(i) for i in SampleList
    )

    if source == "met":
        C_name = "nCG"
    else:
        C_name = "nGC"
    df = pd.DataFrame(
        data_list,
        columns=[
            "source",
            "sample",
            C_name,
            f"{source}_Nmet_reads",
            f"{source}_Nnonmet_reads",
            f"{source}_N_tot_reads",
            f"{source}_rate",
            f"{source}_MT_rate",
            f"{source}_lambda_rate",
            f"{source}_lambda_CpG_count",
            f"{source}_lambda_read_count",
            f"{source}_puc19_rate",
            f"{source}_puc19_CpG_count",
            f"{source}_puc19_read_count",
        ],
    )

    return df


def QC_cpg_and_gpc_counts(
    root_dir,
    source="met",
    force_run=False,
    chunk_size=None,
    cores=64,
):
    """
    root_dir should be the directory to 'acc, met' folders
    """

    file_path = f"{root_dir}/results/DNA_read_counts_{source}.csv"
    df_sample = load_sample_info(root_dir)
    SampleList_0 = df_sample["sample"].to_numpy()
    os.makedirs(f"{root_dir}/results", exist_ok=True)
    print(f"In total {len(set(SampleList_0))} samples")
    SampleList = SampleList_0
    if chunk_size is None:
        chunk_size = len(SampleList)

    if not os.path.exists(file_path) or force_run:
        SampleList = SampleList_0[:chunk_size]
        df = extract_count_statistics(root_dir, SampleList, source=source, cores=cores)
        df.to_csv(file_path, index=0)

    while len(SampleList) > 0:
        df_0 = pd.read_csv(file_path)
        # extract the un-computed samples
        SampleList_1 = SampleList_0[~np.in1d(SampleList_0, df_0["sample"])]
        SampleList = SampleList_1[:chunk_size]
        if len(SampleList) > 0:
            print(f"Remaining {len(SampleList_1)} samples")
            df_1 = extract_count_statistics(
                root_dir, SampleList, source=source, cores=cores
            )
            df = pd.concat([df_0, df_1], ignore_index=True)
        else:
            df = df_0
        df.to_csv(file_path, index=0)
    return df


def plot_lambda_MT_QC(df_info):
    df_info = df_info.sort_values("sample").reset_index()
    for selected_Chr in ["lambda", "MT"]:
        f, axs = plt.subplots(
            1, 2, figsize=(10, 4), gridspec_kw=dict(width_ratios=[4, 4])
        )

        if f"acc_{selected_Chr}_rate" in df_info.columns:
            ax = sns.scatterplot(
                x=df_info.index,
                y=df_info[f"acc_{selected_Chr}_rate"],
                label="acc",
                ax=axs[0],
            )
        ax = sns.scatterplot(
            x=df_info.index,
            y=df_info[f"met_{selected_Chr}_rate"],
            label="met",
            ax=axs[0],
        )
        ax.set_xlabel("Sample order by plate ID")
        ax.set_ylabel(f"Rate ({selected_Chr})")

        if "acc_rate" in df_info.columns:
            ax = sns.scatterplot(
                x=df_info.index,
                y=df_info["acc_rate"],
                label="acc",
                ax=axs[1],
            )
        ax = sns.scatterplot(
            x=df_info.index,
            y=df_info["met_rate"],
            label="met",
            ax=axs[1],
        )
        ax.set_xlabel("Sample order by plate ID")
        ax.set_ylabel("Rate (All chromosomes)")
        f.suptitle(f"{selected_Chr} DNA")
        plt.tight_layout()



import numpy as np
import pandas as pd
import scipy.stats as stats
import seaborn as sns
from ete3 import Tree
from matplotlib import pyplot as plt
from tqdm import tqdm

from . import lineage, metric
from .settings import *

#########################
# clone identification
#########################


def identify_putative_clones_from_trees_with_support(
    tree_path,
    X_similarity,
    support_threshold=0.6,
    print_clone=False,
    similarity_percentile=75,
    verbose=True,
    method="high",
    t_test_threshold_high=1.5,
    t_test_threshold_low=0.75,
    t_test_size_cutoff=3,
    p_value_threshold=0.05,
):
    """
    low: the original method, based on support value and similarity threshold. Seach starting from root.
    This method typically results in low clone resolution.

    high: on top of the low method, further explore whether the identified sub-tree can be furtehr partitioned into sub-clones. When the a sub-tree within the current node can be split into sub-clones, start from this sub-tree, and then
    also add other branches not leading towards this sub-tree.
    This method is currently our best method.

    t-test: on top of the low method, a combination of t-test and support value threshold method to further refine the sub-clones. In t-test, we test whether the current two branches under this node actually belong to two different clones.

    deep: on top of the low method, further search within the current clone for sub-trees that has the right support value. Different from the 'high' method, it does not require two branches under the same node to be valid sub-trees.
    This method tends to produce very small clones.

    shapiro: based on shapiro test
    """

    print(f"method {method}")
    available_methoes = ["high", "low", "t-test", "deep", "shapiro"]
    if method not in available_methoes:
        raise ValueError(f"method is {method}, but should be among {available_methoes}")
    with open(tree_path, "r") as f:
        my_tree = Tree(f.read(), support=True)
        my_tree.support = 0

    tree_list = []
    support = []
    df_list = []

    lineage.traverse_tree_to_update_dist(
        my_tree, X_similarity, percentile=50, rename_leaf_map=lambda x: x.split("-")[0]
    )
    mean_similarity = np.percentile(X_similarity.flatten(), similarity_percentile)

    ##########################
    ### supporting functions
    ##########################
    def collect_putative_clones(node_tmp):
        tree_list.append(node_tmp)
        leaf_idx = np.array(node_tmp.get_leaf_names())
        df_tmp = pd.DataFrame({"cell_id": leaf_idx})
        df_tmp["predicted_clone"] = f"clone_{len(df_list)}"
        df_tmp["clone_support"] = float(node_tmp.support)
        df_tmp["clone_weight"] = float(node_tmp.dist)
        df_tmp["clone_size"] = len(leaf_idx)
        df_list.append(df_tmp)
        if print_clone:  # and (len(leaf_idx)>1):
            renamed_node = node_tmp.copy()
            print(renamed_node)

    def collect_sub_tree_information(node_tmp, level=0, data_list=[], name="0"):
        valid_tmp = (
            (float(node_tmp.support) >= support_threshold)
            & (node_tmp.dist >= mean_similarity)
            & (~node_tmp.is_leaf())
        )
        data_list.append([level, int(valid_tmp), len(node_tmp.get_leaf_names()), name])
        for j0, child_tmp in enumerate(node_tmp.children):
            collect_sub_tree_information(
                child_tmp, level + 1, data_list=data_list, name=f"{name},{j0}"
            )

    def add_clone_info_from_target_level(
        node_tmp, target_level, target_name_list, level=0, name="0"
    ):
        valid_list = []
        invalid_list = []
        for target_name in target_name_list:
            # this sub-tree matches our target names
            condition_1 = target_name == name
            # this sub-tree is along the branch that leads to the target sub-tree, therefore acceptable
            condition_2 = target_name.startswith(name)
            # this sub-tree is contained within one of the target sub-trees
            condition_3 = name.startswith(target_name)
            valid_list.append(condition_1)
            invalid_list.append(condition_2 or condition_3)
        valid = np.sum(valid_list) > 0
        if not valid:
            valid = np.sum(invalid_list) == 0

        if valid:
            collect_putative_clones(node_tmp)
            if name not in target_name_list:
                target_name_list.append(name)

        # print(target_name_list)
        if level < target_level:
            for j0, child_tmp in enumerate(node_tmp.children):
                add_clone_info_from_target_level(
                    child_tmp,
                    target_level,
                    target_name_list,
                    level=level + 1,
                    name=f"{name},{j0}",
                )
        else:
            return

    def refine_clone_based_on_support_high_method(node):
        data_list = []
        collect_sub_tree_information(node, level=0, data_list=data_list)
        df_tmp_orig = pd.DataFrame(
            np.array(data_list), columns=["level", "score", "cell_N", "name"]
        )
        df_tmp_orig["level"] = df_tmp_orig["level"].astype(int)
        df_tmp_orig["score"] = df_tmp_orig["score"].astype(float)
        df_tmp_orig["cell_N"] = df_tmp_orig["cell_N"].astype(int)

        # print(df_tmp_orig)
        df_tmp = (
            df_tmp_orig.groupby("level")
            .agg({"score": "mean", "cell_N": "sum"})
            .reset_index()
        )
        df_tmp = df_tmp[(df_tmp["score"] == 1)]
        # print(df_tmp)
        target_level = df_tmp["level"].max()
        target_name_list = df_tmp_orig[df_tmp_orig["level"] == target_level][
            "name"
        ].to_list()
        add_clone_info_from_target_level(node, target_level, target_name_list)

    def support_value_status(node_tmp, support_threshold, mean_similarity):
        valid_list = []
        for child_tmp in node_tmp.children:
            valid_tmp = (
                (float(child_tmp.support) >= support_threshold)
                & (child_tmp.dist > mean_similarity)
                & (~child_tmp.is_leaf())
            )
            valid_list.append(valid_tmp)
        return np.mean(np.array(valid_list).astype(int)) > 0.999

    def refine_clones_based_on_t_test_and_support(node):

        def t_test_status(node, size_cutoff):
            # a size_cutoff to avoid splitting the clusters to be too small
            child_list = []
            for child in node.children:
                child_list.append(child)
            order_0 = np.array(
                [x.split("-")[0] for x in child_list[0].get_leaf_names()]
            ).astype(int)
            order_1 = np.array(
                [x.split("-")[0] for x in child_list[1].get_leaf_names()]
            ).astype(int)
            if (len(order_0) >= size_cutoff) & (len(order_1) >= size_cutoff):
                X_01 = X_similarity[order_0][:, order_1]
                X_0 = X_similarity[order_0][:, order_0]
                X_1 = X_similarity[order_1][:, order_1]
                tmp_intra = np.hstack(
                    (
                        X_0[np.triu_indices_from(X_0, k=1)],
                        X_1[np.triu_indices_from(X_1, k=1)],
                    )
                )  # [child_list[0].dist, child_list[1].dist]
                tmp_cross = X_01.flatten()

                return (
                    np.mean(tmp_intra) - tmp_cross.mean()
                ) / tmp_intra.std()  # tmp_cross.std()
            else:
                return None

        condition_1 = support_value_status(node, support_threshold, mean_similarity)
        t_test_value = t_test_status(node, t_test_size_cutoff)
        if t_test_value is not None:
            condition_2 = t_test_value > t_test_threshold_high
            condition_3 = t_test_value < t_test_threshold_low
        else:
            condition_2 = False  # not informative
            condition_3 = False  # do not affect whether to split clones or not

        print(
            f"T-test value: {t_test_value } condition {condition_1}   {condition_2} {condition_3}"
        )

        if (condition_1 or condition_2) and (not condition_3):
            for child in node.children:
                refine_clones_based_on_t_test_and_support(child)
        else:
            collect_putative_clones(node)

    def refine_clones_based_on_shapiro_test_and_support(node):
        def shapiro_test(node):
            order_x = np.array([x.split("-")[0] for x in node.get_leaf_names()]).astype(
                int
            )
            if len(order_x) >= t_test_size_cutoff:
                X = X_similarity[order_x][:, order_x]
                X_flat = X[np.triu_indices_from(X, k=1)].flatten()
                result = stats.shapiro(X_flat)
                return result.pvalue
            else:
                return None

        condition_1 = support_value_status(node, support_threshold, mean_similarity)
        p_value = shapiro_test(node)

        if p_value is not None:
            condition_2 = p_value < p_value_threshold
            if condition_2:
                print(f"p_value: {p_value}")
        else:
            condition_2 = False  # not informative

        if condition_1 or condition_2:
            for child in node.children:
                refine_clones_based_on_shapiro_test_and_support(child)
        else:
            collect_putative_clones(node)

    def refine_clones_based_on_support_deep_method(node):
        if (
            (float(node.support) >= support_threshold)
            & (node.dist > mean_similarity)
            & (~node.is_leaf())
        ):
            for child in node.children:
                refine_clones_based_on_support_deep_method(child)
        else:
            collect_putative_clones(node)

    def get_support_distribution(node):
        if not node.is_leaf():
            support.append(float(node.support))
        for child in node.children:
            get_support_distribution(child)

    def identify_clone_main(node):
        if node.is_leaf():
            collect_putative_clones(node)
        elif (
            (float(node.support) >= support_threshold)
            & (node.dist > mean_similarity)
            & (~node.is_leaf())
        ):
            # if print_clone:
            #     print(f'----------full-sub tree-----------')
            #     renamed_node = node.copy()
            #     print(renamed_node)
            #     print('******** collected tree *********')

            if method == "low":
                collect_putative_clones(node)
            elif method == "high":
                refine_clone_based_on_support_high_method(node)
            elif method == "t-test":
                refine_clones_based_on_t_test_and_support(node)
            elif method == "shapiro":
                refine_clones_based_on_shapiro_test_and_support(node)
            else:
                refine_clones_based_on_support_deep_method(node)

        else:
            for child in node.children:
                identify_clone_main(child)

    ###################################
    ####### Call these functions
    ###################################

    identify_clone_main(my_tree.copy())

    df_predict = pd.concat(df_list, ignore_index=True)

    ## rename clone id according to its size
    clone_key = "predicted_clone"
    df_tmp = (
        df_predict.groupby(clone_key)
        .agg({"clone_size": "mean"})
        .sort_values("clone_size", ascending=False)
        .reset_index()
        .reset_index()
    )
    df_tmp[f"new_{clone_key}"] = "clone_" + df_tmp["index"].astype(str)
    clone_key_map = dict(zip(df_tmp[clone_key], df_tmp[f"new_{clone_key}"]))
    df_predict[f"{clone_key}"] = df_predict[clone_key].map(clone_key_map)

    if verbose:
        get_support_distribution(my_tree.copy())
        plt.subplots()
        sns.histplot(support, binwidth=0.05)
        plt.xlabel("Support")
        plt.ylabel("Count")
        df_predict_tmp = df_predict[df_predict["clone_size"] > 1]
        print(
            "Total predicted clones (>1 cells)",
            len(df_predict_tmp["predicted_clone"].unique()),
        )

    return df_predict


def identify_putative_clones(
    X_similarity_old,
    my_tree,
    weight_threshold,
    plot_signal_noise=True,
    rename_leaf_map=None,
    print_clone=False,
):
    ########################################################
    # Signal-noise plot to decide the weight threshold to call clones
    ########################################################
    order_x = np.array(my_tree.get_leaf_names()).astype(int)
    X_similarity = X_similarity_old.copy()

    if plot_signal_noise:
        for i in range(X_similarity.shape[0]):
            X_similarity[i, i] = np.nan  # to avoid counting the diagonal terms

        # Obtain randomized similarity
        X_similarity_rand = X_similarity.copy()
        rand_dist = []
        cell_orders = np.arange(X_similarity.shape[0])
        for __ in tqdm(range(100)):
            np.random.shuffle(cell_orders)
            lineage.traverse_tree_to_update_dist(
                my_tree, X_similarity_rand[cell_orders][:, cell_orders], percentile=50
            )
            for tmp_tree in my_tree.traverse():
                rand_dist.append(tmp_tree.dist)

        # Obtain true similarity
        lineage.traverse_tree_to_update_dist(my_tree, X_similarity, percentile=50)
        true_dist = []
        for tmp_tree in my_tree.traverse():
            true_dist.append(tmp_tree.dist)

        rand_dist = np.array(rand_dist)
        rand_dist = rand_dist[~np.isnan(rand_dist)]
        true_dist = np.array(true_dist)
        true_dist = true_dist[~np.isnan(true_dist)]

        plt.subplots(1, 1, figsize=(5.5, 3.5))
        color_random = "#a6bddb"
        color_data = "#fdbb84"

        all_data = list(true_dist) + list(rand_dist)
        bins = np.linspace(np.min(all_data), np.max(all_data), 50)

        ax = sns.histplot(
            data=true_dist,
            label="Observed",
            bins=bins,
            stat="probability",
            color=color_data,
            alpha=0.5,
        )

        ax = sns.histplot(
            data=rand_dist,
            label="Random",
            bins=bins,
            stat="probability",
            color=color_random,
        )

        # ax.legend()
        plt.legend(loc=[1.05, 0.4])
        # ax.set_xlabel("Sister-cell distance")
        plt.xlabel("Tree weight")
        plt.ylabel("Normalized frequency")
        # plt.yscale('log')
        plt.tight_layout()

    ################## Actually identify clones
    lineage.traverse_tree_to_update_dist(my_tree, X_similarity, percentile=50)
    tree_list = []
    dist = []
    df_list = []

    def traverse_tree(node):
        dist.append(node.dist)
        if node.dist > weight_threshold:
            if print_clone and rename_leaf_map is not None:
                renamed_node = node.copy()
                for j, leaf in enumerate(renamed_node.iter_leaves()):
                    # Modify the leaf name based on your desired logic
                    new_name = f"{j}" + "; " + str(rename_leaf_map[leaf.name])
                    leaf.name = new_name
                print(renamed_node)

            tree_list.append(node)
            leaf_idx = np.array(node.get_leaf_names()).astype(int)
            df_tmp = pd.DataFrame({"cell_id": leaf_idx})
            df_tmp["predicted_clone"] = f"clone_{len(df_list)}"
            df_tmp["clone_weight"] = node.dist
            df_tmp["clone_size"] = len(leaf_idx)
            df_list.append(df_tmp)

        else:
            for child in node.children:
                traverse_tree(child)

    traverse_tree(my_tree.copy())
    # print("Total count", len(tree_list))

    df_predict = pd.concat(df_list, ignore_index=True)
    df = (
        pd.DataFrame({"cell_id": order_x})
        .merge(df_predict, on="cell_id", how="outer")
        .sort_values("cell_id")
        .reset_index(drop=True)
    )
    return df.sort_values("clone_weight", ascending=False)


def infer_clone_number_across_various_threshold(
    adata, out_dir, clone_key, save_data_des, fig_dir
):
    my_tree_path = f"{out_dir}/lineage_tree_{clone_key}_{save_data_des}.txt"
    df_state_clone = pd.read_csv(
        f"{out_dir}/state_clone_info_{clone_key}_{save_data_des}.csv"
    )
    rename_leaf_map = dict(
        zip(df_state_clone["state_info"].astype(str), df_state_clone["clone_id"])
    )

    with open(my_tree_path, "r") as f:
        my_tree = Tree(f.read(), support=False)

    order_x = np.array(my_tree.get_leaf_names()).astype(int)

    used_threshold_list = []
    accuracy_list = []
    fraction_list = []
    obs_clone_N_list = []
    estimated_total_clone_N_list = []
    for infer_clone_threshold_tmp in np.arange(0.4, 0.9, 0.02):
        df_tmp = identify_putative_clones(
            adata.obsm["X_similarity"],
            my_tree.copy(),
            weight_threshold=infer_clone_threshold_tmp,
            plot_signal_noise=False,
            print_clone=False,
            rename_leaf_map=rename_leaf_map,
        )

        adata.obs["inferred_clones"] = df_tmp.sort_values("cell_id")[
            "predicted_clone"
        ].to_numpy()

        sel_idx = pd.isna(adata.obs["inferred_clones"])
        old_clone_array = np.array(adata.obs[clone_key].copy()).astype(str)
        old_clone_array[sel_idx] = "nan"

        df_accuracy = metric.lineage_accuracy_from_leaf_order(old_clone_array[order_x])
        obs_clone_N = len(df_tmp["predicted_clone"].unique())
        estimated_total_clone_N = estimate_clone_number(
            adata,
            plot=False,
            fig_dir=fig_dir,
            data_des=f"Thresh: {infer_clone_threshold_tmp:.2f}",
        )

        if len(df_accuracy) > 0:
            used_threshold_list.append(infer_clone_threshold_tmp)
            accuracy_list.append(df_accuracy["accuracy"].mean())
            fraction_list.append(np.sum(~sel_idx) / len(sel_idx))
            obs_clone_N_list.append(obs_clone_N)
            estimated_total_clone_N_list.append(estimated_total_clone_N)

    df_clone_test = pd.DataFrame(
        {
            "threshold": used_threshold_list,
            "accuracy": accuracy_list,
            "retained_cell_fraction": fraction_list,
            "observed_clone_N": obs_clone_N_list,
            "Estimated_total_clone_N": estimated_total_clone_N_list,
        }
    )
    plt.subplots(figsize=(4, 3))
    sns.scatterplot(data=df_clone_test, x="threshold", y="accuracy")
    plt.xlabel("Weight threshold")
    plt.ylabel("Lineage accuracy")
    plt.subplots(figsize=(4, 3))
    sns.scatterplot(data=df_clone_test, x="threshold", y="retained_cell_fraction")
    plt.xlabel("Weight threshold")
    plt.ylabel("Retained cell fraction")

    plt.subplots(figsize=(4, 3))
    plt.plot(
        df_clone_test["threshold"],
        df_clone_test["observed_clone_N"],
        "*",
        label="Obs_N",
    )
    plt.plot(
        df_clone_test["threshold"],
        df_clone_test["Estimated_total_clone_N"],
        "o",
        label="Esti_Total_N",
    )
    plt.xlabel("Threshold")
    plt.ylabel("Clone number")
    plt.legend()


def estimate_clone_number_from_DARLIN_barcodes(
    pd_series, plot=True, clone_key="inferred_clones"
):
    df = (
        pd.DataFrame({clone_key: list(pd_series.astype(str))})
        .reset_index()
        .groupby(clone_key)
        .agg({"index": "count"})
        .rename(columns={"index": "clone_size"})
    )
    clone_size_stat = {}

    for j in range(1, df[~df.index.isin(["nan"])]["clone_size"].max() + 1):
        clone_size_stat[j] = np.sum(df["clone_size"] == j)

    ## a clone of size 2 is counted as a clone of size 1
    # for j in range(2, df[~df.index.isin(["nan"])]["clone_size"].max() + 1):
    #     clone_size_stat[j - 1] = np.sum(df["clone_size"] == j)

    df_freq = pd.DataFrame(
        {"clone_size": clone_size_stat.keys(), "frequency": clone_size_stat.values()}
    )

    tot_clones = df_freq["frequency"].sum()
    singleton_N = df_freq[df_freq["clone_size"] == 1]["frequency"].values[0]
    singleton_ratio = singleton_N / tot_clones
    doublet_N = df_freq[df_freq["clone_size"] == 2]["frequency"].values[0]
    doublet_ratio = doublet_N / tot_clones

    # Good-Turing estimator
    # inferred_clone_N_1 = tot_clones / (1 - singleton_ratio)

    # Chao1 estimator
    inferred_clone_N_2 = tot_clones + singleton_N**2 / (2 * doublet_N)

    if plot:
        sns.scatterplot(data=df_freq, x="clone_size", y="frequency")
        plt.title(
            f"Obs clone={tot_clones}"
            + "\n"
            + f"Estimated clone_N (Chao1)={inferred_clone_N_2:.0f}"
        )
        plt.xlabel("Clone size")
        plt.ylabel("Frequency")
    return inferred_clone_N_2


def estimate_clone_number(adata, clone_key="inferred_clones", nan_as_size_1=False):
    df = (
        pd.DataFrame({clone_key: adata.obs[clone_key].astype(str).to_list()})
        .reset_index()
        .groupby(clone_key)
        .agg({"index": "count"})
        .rename(columns={"index": "clone_size"})
    )
    clone_size_stat = {}

    if nan_as_size_1:
        # if we set the threshold high, then most of cells will be 'nan', thus inflating the estimation
        clone_size_stat[1] = df.loc["nan"]["clone_size"]
        for j in range(2, df[~df.index.isin(["nan"])]["clone_size"].max() + 1):
            clone_size_stat[j] = np.sum(df["clone_size"] == j)
    else:
        # # a clone of size 2 is counted as a clone of size 1
        # for j in range(2, df[~df.index.isin(["nan"])]["clone_size"].max() + 1):
        #     clone_size_stat[j - 1] = np.sum(df["clone_size"] == j)

        # a clone of size 2 is counted as a clone of size 2
        for j in range(1, df[~df.index.isin(["nan"])]["clone_size"].max() + 1):
            clone_size_stat[j] = np.sum(df["clone_size"] == j)

    df_freq = pd.DataFrame(
        {"clone_size": clone_size_stat.keys(), "frequency": clone_size_stat.values()}
    )

    tot_clones = df_freq["frequency"].sum()
    singleton_N = df_freq[df_freq["clone_size"] == 1]["frequency"].values[0]
    singleton_ratio = singleton_N / tot_clones
    if len(df_freq[df_freq["clone_size"] == 2]["frequency"].values) == 0:
        return np.nan
    else:
        doublet_N = df_freq[df_freq["clone_size"] == 2]["frequency"].values[0]
        doublet_ratio = doublet_N / tot_clones

    # Good-Turing estimator
    # inferred_clone_N_1 = tot_clones / (1 - singleton_ratio)

    # Chao1 estimator
    inferred_clone_N_2 = tot_clones + singleton_N**2 / (2 * doublet_N)

    print(f"Predicted total clones by Chao1: {inferred_clone_N_2:.0f}")

    frac_cell_from_singleton_clone = (
        df_freq[df_freq["clone_size"] == 1].frequency.values[0] / adata.shape[0]
    )

    print(
        f"Fraction of cells from singleton clones: {frac_cell_from_singleton_clone:.2f}"
    )

    return df_freq
